import math
import random
from enum import Enum

REAL_EPS = 1e-13


class BitEncoding(Enum):
    """Bit Encoding"""

    UNSIGNED = 0
    TWOS_COMPLEMENT = 1


class PhaseFunc(Enum):
    """PhaseFunc"""

    NORM = 0
    SCALED_NORM = 1
    INVERSE_NORM = 2
    SCALED_INVERSE_NORM = 3
    SCALED_INVERSE_SHIFTED_NORM = 4
    PRODUCT = 5
    SCALED_PRODUCT = 6
    INVERSE_PRODUCT = 7
    SCALED_INVERSE_PRODUCT = 8
    DISTANCE = 9
    SCALED_DISTANCE = 10
    INVERSE_DISTANCE = 11
    SCALED_INVERSE_DISTANCE = 12
    SCALED_INVERSE_SHIFTED_DISTANCE = 13


class PauliOpType(Enum):
    """PauliOpType"""

    PAULI_I = 0
    PAULI_X = 1
    PAULI_Y = 2
    PAULI_Z = 3


class Simulator:
    """Simulator implement"""

    def __init__(self):
        self.real = []  # real
        self.imag = []  # imag
        self.qureg = None
        self.qubits = 0
        self.total_num_amps = 0  # numAmpsPerChunk

    def create_qureg(self, num_qubits):
        """Allocate resource.

        Args:
            num_qubits: number of qubits
        """
        self.qubits = num_qubits
        num_amps = 2**num_qubits
        self.total_num_amps = num_amps
        # TODO: need to improve.
        self.real = [0] * num_amps
        self.imag = [0] * num_amps

    def init_blank_state(self):
        """Init blank state"""
        for i in range(self.total_num_amps):
            self.real[i] = 0.0
            self.imag[i] = 0.0

    def init_zero_state(self):
        """Init zero state"""
        self.init_blank_state()
        # TODO:??
        self.real[0] = 1.0

    def init_plus_state(self):
        """Init plus state"""
        # dimension of the state vector
        chunk_size = self.total_num_amps
        state_vec_size = chunk_size
        normFactor = 1.0 / math.sqrt(state_vec_size)

        # initialise the state to |+++..+++> = 1/normFactor {1, 1, 1, ...}
        for this_task in range(chunk_size):
            self.real[this_task] = normFactor
            self.imag[this_task] = 0.0

    def init_classical_state(self):
        """Init classical state"""

        # initialise the state to vector to all zeros
        self.init_blank_state()

        # give the specified classical state prob 1
        self.real[0] = 1.0
        self.imag[0] = 0.0

    def amp(self, reals, imags, startindex, numamps):
        """Init amplitudes state"""
        
        for index in range(numamps):
            self.real[startindex] = reals[index]
            self.imag[startindex] = imags[index]
            startindex += 1

    def matrix(self, controls, targets, reals, imags):
        """Apply custom matrix"""
        ctrlMask = self.get_qubit_bit_mask(controls, len(controls))

        numTargs = len(targets)
        numTasks = self.total_num_amps >> numTargs
        numTargAmps = 1 << numTargs

        ampInds = [0]*numTargAmps
        reAmps = [0]*numTargAmps
        imAmps = [0]*numTargAmps

        sortedTargs = [0]*numTargs
        for t in range(numTargs):
            sortedTargs[t] = targets[t]
        sortedTargs.sort()

        for thisTask in range(numTasks):
            # find this task's start index (where all targs are 0)
            thisInd00 = thisTask
            for t in range(numTargs):
                thisInd00 = self.insert_zero_bit(thisInd00, sortedTargs[t])
                
            # this task only modifies amplitudes if control qubits are 1 for this state
            thisGlobalInd00 = thisInd00
            if (ctrlMask and ((ctrlMask & thisGlobalInd00) != ctrlMask)):
                continue
                
            # determine the indices and record values of this tasks's target amps
            for i in range(numTargAmps):
                # get statevec index of current target qubit assignment
                ind = thisInd00
                for t in range(numTargs):
                    if (self.extract_bit(t, i)):
                        ind = self.flip_bit(ind, targets[t])
                
                # update this tasks's private arrays
                ampInds[i] = ind
                reAmps [i] = self.real[ind]
                imAmps [i] = self.imag[ind]
            
            # modify this tasks's target amplitudes
            for r in range(numTargAmps):
                ind = ampInds[r]
                self.imag[ind] = 0
                self.imag[ind] = 0
                
                for c in range(numTargAmps):
                    reElem = reals[r][c]
                    imElem = imags[r][c]
                    self.real[ind] += reAmps[c]*reElem - imAmps[c]*imElem
                    self.imag[ind] += reAmps[c]*imElem + imAmps[c]*reElem

    # TODO:need to improve.
    def hadamard(self, target):
        """
        Apply hadamard gate.

        Args:
            target: target qubit.
        """
        size_half_block = 2**target
        size_block = size_half_block * 2
        num_task = self.total_num_amps // 2

        rec_root = 1.0 / math.sqrt(2)
        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            state_real_up = self.real[index_up]
            state_imag_up = self.imag[index_up]

            state_real_lo = self.real[index_lo]
            state_imag_lo = self.imag[index_lo]

            self.real[index_up] = rec_root * (state_real_up + state_real_lo)
            self.imag[index_up] = rec_root * (state_imag_up + state_imag_lo)

            self.real[index_lo] = rec_root * (state_real_up - state_real_lo)
            self.imag[index_lo] = rec_root * (state_imag_up - state_imag_lo)

    def ch(self, control_bit, target_bit, ureal, uimage):
        self.controlled_unitary(control_bit, target_bit, ureal, uimage)

    def phase_shift(self, target, angle):
        """Shift the phase between |0> and |1> of a single qubit by a given angle.

        Args:
            target: qubit to undergo a phase shift.
            angle:  amount by which to shift the phase in radians.
        """
        real = math.cos(angle)
        imag = math.sin(angle)
        self.phase_shift_by_term(self.real, self.imag, target, real, imag)

    def controlled_phase_shift(self, ctrl, target, angle):
        """
        Controlled-Phase gate.

        Args:
            ctrl: control qubit
            target: target qubit
            angle: amount by which to shift the phase in radians.
        """
        state_vec_size = self.total_num_amps
        cos_angle = math.cos(angle)
        sin_angle = math.sin(angle)

        for index in range(state_vec_size):
            bit1 = self.extract_bit(ctrl, index)
            bit2 = self.extract_bit(target, index)

            if bit1 and bit2:
                state_real_lo = self.real[index]
                state_imag_lo = self.imag[index]

                self.real[index] = cos_angle * state_real_lo - sin_angle * state_imag_lo
                self.imag[index] = sin_angle * state_real_lo + cos_angle * state_imag_lo

    def rotate(self, target, ureal, uimag):
        """Rotate gate."""
        self.apply_matrix2(target, ureal, uimag)

    def rotate_x(self, target, angle):
        """rx gate."""
        unit_axis = [1, 0, 0]
        self.rotate_around_axis(target, angle, unit_axis)

    def rotate_y(self, target, angle):
        """ry gate."""
        unit_axis = [0, 1, 0]
        self.rotate_around_axis(target, angle, unit_axis)

    def rotate_z(self, target, angle):
        """rz gate."""
        unit_axis = [0, 0, 1]
        self.rotate_around_axis(target, angle, unit_axis)

    def pauli_x(self, target):
        """The single-qubit Pauli-X gate."""
        self.paulix_local(self.real, self.imag, target)

    def pauli_y(self, target):
        """The single-qubit Pauli-Y gate."""
        conj_fac = 1
        self.pauliy_local(self.real, self.imag, target, conj_fac)

    def pauli_z(self, target):
        """The single-qubit Pauli-Z gate."""
        real = -1
        imag = 0
        self.phase_shift_by_term(self.real, self.imag, target, real, imag)

    def s_gate(self, target):
        """The single-qubit S gate."""
        real = 0
        imag = 1
        self.phase_shift_by_term(self.real, self.imag, target, real, imag)

    def t_gate(self, target):
        """The single-qubit T gate."""
        real = 1 / math.sqrt(2)
        imag = -1 / math.sqrt(2)
        self.phase_shift_by_term(self.real, self.imag, target, real, imag)

    def control_not(self, ctrl, target):
        """Control not gate"""
        num_task = self.total_num_amps // 2
        size_half_block = 2**target
        size_block = size_half_block * 2

        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            control_bit = self.extract_bit(ctrl, index_up)
            if control_bit:
                state_real_up = self.real[index_up]
                state_imag_up = self.imag[index_up]

                self.real[index_up] = self.real[index_lo]
                self.imag[index_up] = self.imag[index_lo]

                self.real[index_lo] = state_real_up
                self.imag[index_lo] = state_imag_up

    def sdg(self, target_bit, ureal, uimag):
        self.apply_matrix2(target_bit, ureal, uimag)

    def tdg(self, target_bit, ureal, uimag):
        self.apply_matrix2(target_bit, ureal, uimag)

    def sqrtswap(self, target_bit0, target_bit1, ureal, uimag):
        self.apply_matrix4(target_bit0, target_bit1, ureal, uimag)

    def swap(self, target_bit0, target_bit1):
        num_task = self.total_num_amps >> 2
        for this_task in range(num_task):
            # determine ind00 of |..0..0..>, |..0..1..> and |..1..0..>
            ind00 = self.insert_two_zero_bits(this_task, target_bit0, target_bit1)
            ind01 = self.flip_bit(ind00, target_bit0)
            ind10 = self.flip_bit(ind00, target_bit1)

            # extract statevec amplitudes
            re01 = self.real[ind01]
            im01 = self.imag[ind01]
            re10 = self.real[ind10]
            im10 = self.imag[ind10]

            # swap 01 and 10 amps
            self.real[ind01] = re10
            self.real[ind10] = re01
            self.imag[ind01] = im10
            self.imag[ind10] = im01

    def cswap(self, control_bit, target_bit0, targetbi_bit1, ureal, uimage):
        self.controlled_two_qubit_unitary(
            control_bit, target_bit0, targetbi_bit1, ureal, uimage
        )

    def cy(self, control_bit, target_bit):
        conj_fac = 1
        num_task = self.total_num_amps >> 1
        size_half_block = 2**target_bit
        size_block = size_half_block * 2
        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            controlBit = self.extract_bit(control_bit, index_up)
            if controlBit:
                state_real_up = self.real[index_up]
                state_imag_up = self.imag[index_up]

                # update under +-{{0, -i}, {i, 0}}
                self.real[index_up] = conj_fac * self.imag[index_lo]
                self.imag[index_up] = conj_fac * -self.real[index_lo]
                self.real[index_lo] = conj_fac * -state_imag_up
                self.imag[index_lo] = conj_fac * state_real_up

    def cz(self, control_bits, num_control_bits):
        state_vec_size = self.total_num_amps
        mask = self.get_qubit_bit_mask(control_bits, num_control_bits)
        for index in range(state_vec_size):
            if mask == (mask & index):
                self.real[index] = -self.real[index]
                self.imag[index] = -self.imag[index]

    def u3(self, target_bit, ureal, uimag):
        self.unitary(target_bit, ureal, uimag)

    def u2(self, target_bit, ureal, uimag):
        self.unitary(target_bit, ureal, uimag)

    def u1(self, target_bit, ureal, uimag):
        self.unitary(target_bit, ureal, uimag)

    def unitary(self, target_bit, ureal, uimag):
        num_task = self.total_num_amps >> 1
        size_half_block = 2**target_bit
        size_block = size_half_block * 2

        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            # store current state vector values in temp variables
            state_real_up = self.real[index_up]
            state_imag_up = self.imag[index_up]

            state_real_lo = self.real[index_lo]
            state_imag_lo = self.imag[index_lo]

            self.real[index_up] = (
                ureal[0][0] * state_real_up
                - uimag[0][0] * state_imag_up
                + ureal[0][1] * state_real_lo
                - uimag[0][1] * state_imag_lo
            )
            self.imag[index_up] = (
                ureal[0][0] * state_imag_up
                + uimag[0][0] * state_real_up
                + ureal[0][1] * state_imag_lo
                + uimag[0][1] * state_real_lo
            )

            self.real[index_lo] = (
                ureal[1][0] * state_real_up
                - uimag[1][0] * state_imag_up
                + ureal[1][1] * state_real_lo
                - uimag[1][1] * state_imag_lo
            )
            self.imag[index_lo] = (
                ureal[1][0] * state_imag_up
                + uimag[1][0] * state_real_up
                + ureal[1][1] * state_imag_lo
                + uimag[1][1] * state_real_lo
            )

    def crx(self, control_bit, target_bit, angle):
        unit_axis = [1, 0, 0]
        self.controlled_rotate_around_axis(control_bit, target_bit, angle, unit_axis)

    def cry(self, control_bit, target_bit, angle):
        unit_axis = [0, 1, 0]
        self.controlled_rotate_around_axis(control_bit, target_bit, angle, unit_axis)

    def crz(self, control_bit, target_bit, angle):
        unit_axis = [0, 0, 1]
        self.controlled_rotate_around_axis(control_bit, target_bit, angle, unit_axis)

    def controlled_rotate_around_axis(self, control_bit, target_bit, angle, unit_axis):
        mag = math.sqrt(
            unit_axis[0] * unit_axis[0]
            + unit_axis[1] * unit_axis[1]
            + unit_axis[2] * unit_axis[2]
        )
        unit_vec = [unit_axis[0] / mag, unit_axis[1] / mag, unit_axis[2] / mag]

        alpha_real = math.cos(angle / 2.0)
        alpha_imag = -math.sin(angle / 2.0) * unit_vec[2]
        beta_real = math.sin(angle / 2.0) * unit_vec[1]
        beta_imag = -math.sin(angle / 2.0) * unit_vec[0]

        num_task = self.total_num_amps >> 1
        size_half_block = 2**target_bit
        size_block = size_half_block * 2
        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            control_bit = self.extract_bit(control_bit, index_up)
            if control_bit:
                # store current state vector values in temp variables
                state_real_up = self.real[index_up]
                state_imag_up = self.imag[index_up]

                state_real_lo = self.real[index_lo]
                state_imag_lo = self.imag[index_lo]

                self.real[index_up] = (
                    alpha_real * state_real_up
                    - alpha_imag * state_imag_up
                    - beta_real * state_real_lo
                    - beta_imag * state_imag_lo
                )
                self.imag[index_up] = (
                    alpha_real * state_imag_up
                    + alpha_imag * state_real_up
                    - beta_real * state_imag_lo
                    + beta_imag * state_real_lo
                )

                self.real[index_lo] = (
                    beta_real * state_real_up
                    - beta_imag * state_imag_up
                    + alpha_real * state_real_lo
                    + alpha_imag * state_imag_lo
                )
                self.imag[index_lo] = (
                    beta_real * state_imag_up
                    + beta_imag * state_real_up
                    + alpha_real * state_imag_lo
                    - alpha_imag * state_real_lo
                )

    def rotate_around_axis(self, target_bit, angle, unit_axis):
        mag = math.sqrt(
            unit_axis[0] * unit_axis[0]
            + unit_axis[1] * unit_axis[1]
            + unit_axis[2] * unit_axis[2]
        )
        unit_vec = [unit_axis[0] / mag, unit_axis[1] / mag, unit_axis[2] / mag]

        alpha_real = math.cos(angle / 2.0)
        alpha_imag = -math.sin(angle / 2.0) * unit_vec[2]
        beta_real = math.sin(angle / 2.0) * unit_vec[1]
        beta_imag = -math.sin(angle / 2.0) * unit_vec[0]

        num_task = self.total_num_amps >> 1
        size_half_block = 2**target_bit
        size_block = size_half_block * 2
        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            # store current state vector values in temp variables
            state_real_up = self.real[index_up]
            state_imag_up = self.imag[index_up]

            state_real_lo = self.real[index_lo]
            state_imag_lo = self.imag[index_lo]

            self.real[index_up] = (
                alpha_real * state_real_up
                - alpha_imag * state_imag_up
                - beta_real * state_real_lo
                - beta_imag * state_imag_lo
            )
            self.imag[index_up] = (
                alpha_real * state_imag_up
                + alpha_imag * state_real_up
                - beta_real * state_imag_lo
                + beta_imag * state_real_lo
            )

            self.real[index_lo] = (
                beta_real * state_real_up
                - beta_imag * state_imag_up
                + alpha_real * state_real_lo
                + alpha_imag * state_imag_lo
            )
            self.imag[index_lo] = (
                beta_real * state_imag_up
                + beta_imag * state_real_up
                + alpha_real * state_imag_lo
                - alpha_imag * state_real_lo
            )

    def x1(self, target, ureal, uimag):
        self.apply_matrix2(target, ureal, uimag)

    def y1(self, target, ureal, uimag):
        self.apply_matrix2(target, ureal, uimag)

    def z1(self, target, ureal, uimag):
        self.apply_matrix2(target, ureal, uimag)

    def sqrtx(self, target, ureal, uimag):
        self.apply_matrix2(target, ureal, uimag)

    def sqrtxdg(self, target, ureal, uimag):
        self.apply_matrix2(target, ureal, uimag)

    def csqrtx(self, control_bit, target_bit, ureal, uimage):
        self.controlled_unitary(control_bit, target_bit, ureal, uimage)

    def cu1(self, target_bit0, target_bit1, ureal, uimag):
        self.apply_matrix4(target_bit0, target_bit1, ureal, uimag)

    def cu3(self, target_bit0, target_bit1, ureal, uimag):
        self.apply_matrix4(target_bit0, target_bit1, ureal, uimag)

    def cu(self, target_bit0, target_bit1, ureal, uimag):
        self.apply_matrix4(target_bit0, target_bit1, ureal, uimag)

    def cr(self, target_bit0, target_bit1, ureal, uimag):
        self.apply_matrix4(target_bit0, target_bit1, ureal, uimag)

    def iswap(self, target_bit0, target_bit1, ureal, uimag):
        self.apply_matrix4(target_bit0, target_bit1, ureal, uimag)

    def get_qubit_bit_mask(self, qubits, numqubit):
        mask = 0
        for index in range(numqubit):
            mask = mask | (1 << qubits[index])

        return mask

    def apply_matrix2(self, target, ureal, uimag):
        num_task = self.total_num_amps // 2
        size_half_block = 2**target
        size_block = size_half_block * 2
        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            state_real_up = self.real[index_up]
            state_imag_up = self.imag[index_up]

            state_real_lo = self.real[index_lo]
            state_imag_lo = self.imag[index_lo]

            self.real[index_up] = (
                ureal[0][0] * state_real_up
                - uimag[0][0] * state_imag_up
                + ureal[0][1] * state_real_lo
                - uimag[0][1] * state_imag_lo
            )
            self.imag[index_up] = (
                ureal[0][0] * state_imag_up
                + uimag[0][0] * state_real_up
                + ureal[0][1] * state_imag_lo
                + uimag[0][1] * state_real_lo
            )
            self.real[index_lo] = (
                ureal[1][0] * state_real_up
                - uimag[1][0] * state_imag_up
                + ureal[1][1] * state_real_lo
                - uimag[1][1] * state_imag_lo
            )
            self.imag[index_lo] = (
                ureal[1][0] * state_imag_up
                + uimag[1][0] * state_real_up
                + ureal[1][1] * state_imag_lo
                + uimag[1][1] * state_real_lo
            )

    def apply_matrix4(self, target0, target1, ureal, uimag):
        ctrl_mask = 0
        global_ind_start = self.total_num_amps
        num_task = self.total_num_amps >> 2
        for this_task in range(num_task):
            ind00 = self.insert_two_zero_bits(this_task, target0, target1)
            this_global_ind00 = ind00 + global_ind_start
            if ctrl_mask and ((ctrl_mask & this_global_ind00) != ctrl_mask):
                continue

            # inds of |..0..1..>, |..1..0..> and |..1..1..>
            ind01 = self.flip_bit(ind00, target0)
            ind10 = self.flip_bit(ind00, target1)
            ind11 = self.flip_bit(ind01, target1)

            # extract statevec amplitudes
            re00 = self.real[ind00]
            im00 = self.imag[ind00]
            re01 = self.real[ind01]
            im01 = self.imag[ind01]
            re10 = self.real[ind10]
            im10 = self.imag[ind10]
            re11 = self.real[ind11]
            im11 = self.imag[ind11]

            # apply u * {amp00, amp01, amp10, amp11}
            self.real[ind00] = (
                ureal[0][0] * re00
                - uimag[0][0] * im00
                + ureal[0][1] * re01
                - uimag[0][1] * im01
                + ureal[0][2] * re10
                - uimag[0][2] * im10
                + ureal[0][3] * re11
                - uimag[0][3] * im11
            )
            self.imag[ind00] = (
                uimag[0][0] * re00
                + ureal[0][0] * im00
                + uimag[0][1] * re01
                + ureal[0][1] * im01
                + uimag[0][2] * re10
                + ureal[0][2] * im10
                + uimag[0][3] * re11
                + ureal[0][3] * im11
            )

            self.real[ind01] = (
                ureal[1][0] * re00
                - uimag[1][0] * im00
                + ureal[1][1] * re01
                - uimag[1][1] * im01
                + ureal[1][2] * re10
                - uimag[1][2] * im10
                + ureal[1][3] * re11
                - uimag[1][3] * im11
            )
            self.imag[ind01] = (
                uimag[1][0] * re00
                + ureal[1][0] * im00
                + uimag[1][1] * re01
                + ureal[1][1] * im01
                + uimag[1][2] * re10
                + ureal[1][2] * im10
                + uimag[1][3] * re11
                + ureal[1][3] * im11
            )

            self.real[ind10] = (
                ureal[2][0] * re00
                - uimag[2][0] * im00
                + ureal[2][1] * re01
                - uimag[2][1] * im01
                + ureal[2][2] * re10
                - uimag[2][2] * im10
                + ureal[2][3] * re11
                - uimag[2][3] * im11
            )
            self.imag[ind10] = (
                uimag[2][0] * re00
                + ureal[2][0] * im00
                + uimag[2][1] * re01
                + ureal[2][1] * im01
                + uimag[2][2] * re10
                + ureal[2][2] * im10
                + uimag[2][3] * re11
                + ureal[2][3] * im11
            )

            self.real[ind11] = (
                ureal[3][0] * re00
                - uimag[3][0] * im00
                + ureal[3][1] * re01
                - uimag[3][1] * im01
                + ureal[3][2] * re10
                - uimag[3][2] * im10
                + ureal[3][3] * re11
                - uimag[3][3] * im11
            )
            self.imag[ind11] = (
                uimag[3][0] * re00
                + ureal[3][0] * im00
                + uimag[3][1] * re01
                + ureal[3][1] * im01
                + uimag[3][2] * re10
                + ureal[3][2] * im10
                + uimag[3][3] * re11
                + ureal[3][3] * im11
            )

    def controlled_unitary(self, control_bit, target_bit, ureal, uimag):
        num_task = self.total_num_amps >> 1
        size_half_block = 1 << target_bit
        size_block = 2 * size_half_block

        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            control_bit = self.extract_bit(control_bit, index_up)
            if control_bit:
                # store current state vector values in temp variables.
                state_real_up = self.real[index_up]
                state_imag_up = self.imag[index_up]

                state_real_lo = self.real[index_lo]
                state_imag_lo = self.imag[index_lo]

                self.real[index_up] = (
                    ureal[0][0] * state_real_up
                    - uimag[0][0] * state_imag_up
                    + ureal[0][1] * state_real_lo
                    - uimag[0][1] * state_imag_lo
                )
                self.imag[index_up] = (
                    ureal[0][0] * state_imag_up
                    + uimag[0][0] * state_real_up
                    + ureal[0][1] * state_imag_lo
                    + uimag[0][1] * state_real_lo
                )

                self.real[index_lo] = (
                    ureal[1][0] * state_real_up
                    - uimag[1][0] * state_imag_up
                    + ureal[1][1] * state_real_lo
                    - uimag[1][1] * state_imag_lo
                )
                self.imag[index_lo] = (
                    ureal[1][0] * state_imag_up
                    + uimag[1][0] * state_real_up
                    + ureal[1][1] * state_imag_lo
                    + uimag[1][1] * state_real_lo
                )

    def insert_two_zero_bits(self, number, bit1, bit2):
        small = bit1 if bit1 < bit2 else bit2
        big = bit2 if bit1 < bit2 else bit1
        return self.insert_zero_bit(self.insert_zero_bit(number, small), big)

    def insert_zero_bit(self, number, index):
        left = (number >> index) << index
        right = number - left
        return (left << 1) ^ right

    def flip_bit(self, number, bit_ind):
        return number ^ (1 << bit_ind)

    def extract_bit(self, ctrl, index):
        return (index & (2**ctrl)) // (2**ctrl)

    def multi_controlled_multi_qubit_not(
        self, control_bits, num_control_bits, target_bits, num_target_bits
    ):
        ctrl_mask = self.get_qubit_bit_mask(control_bits, num_control_bits)
        targ_mask = self.get_qubit_bit_mask(target_bits, num_target_bits)

        for amp_ind in range(self.total_num_amps):

            # /* it may be a premature optimisation to remove the seemingly wasteful continues below,
            #  * because the maximum skipped amplitudes is 1/2 that stored in the node
            #  * (e.g. since this function is not called if all amps should be skipped via controls),
            #  * and since we're memory-bandwidth bottlenecked.
            #  */

            # // although amps are local, we may still be running in distributed mode,
            # // and hence need to consult the global index to determine the values of
            # // the control qubits
            global_ind = amp_ind

            # // modify amplitude only if control qubits are 1 for this state
            if ctrl_mask and ((ctrl_mask & global_ind) != ctrl_mask):
                continue

            mate_ind = amp_ind ^ targ_mask

            # // if the mate is behind, it was already processed
            if mate_ind < amp_ind:
                continue

            mate_re = self.real[mate_ind]
            mate_im = self.imag[mate_ind]

            # // swap amp with mate
            self.real[mate_ind] = self.real[amp_ind]
            self.imag[mate_ind] = self.imag[amp_ind]
            self.real[amp_ind] = mate_re
            self.imag[amp_ind] = mate_im

    def controlled_two_qubit_unitary(
        self, control_bit, target_bit0, targetbi_bit1, ureal, uimag
    ):
        num_task = self.total_num_amps >> 2
        for this_task in range(num_task):
            #  determine ind00 of |..0..0..>.
            ind00 = self.insert_two_zero_bits(this_task, target_bit0, targetbi_bit1)

            #  skip amplitude if controls aren't in 1 state (overloaded for speed).
            this_global_ind00 = ind00
            if control_bit and ((control_bit & this_global_ind00) != control_bit):
                continue

            #  inds of |..0..1..>, |..1..0..> and |..1..1..>.
            ind01 = self.flip_bit(ind00, target_bit0)
            ind10 = self.flip_bit(ind00, targetbi_bit1)
            ind11 = self.flip_bit(ind01, targetbi_bit1)

            #  extract statevec amplitudes
            re00 = self.real[ind00]
            im00 = self.imag[ind00]
            re01 = self.real[ind01]
            im01 = self.imag[ind01]
            re10 = self.real[ind10]
            im10 = self.imag[ind10]
            re11 = self.real[ind11]
            im11 = self.imag[ind11]

            #  apply u * {amp00, amp01, amp10, amp11}.
            self.real[ind00] = (
                ureal[0][0] * re00
                - uimag[0][0] * im00
                + ureal[0][1] * re01
                - uimag[0][1] * im01
                + ureal[0][2] * re10
                - uimag[0][2] * im10
                + ureal[0][3] * re11
                - uimag[0][3] * im11
            )
            self.imag[ind00] = (
                uimag[0][0] * re00
                + ureal[0][0] * im00
                + uimag[0][1] * re01
                + ureal[0][1] * im01
                + uimag[0][2] * re10
                + ureal[0][2] * im10
                + uimag[0][3] * re11
                + ureal[0][3] * im11
            )

            self.real[ind01] = (
                ureal[1][0] * re00
                - uimag[1][0] * im00
                + ureal[1][1] * re01
                - uimag[1][1] * im01
                + ureal[1][2] * re10
                - uimag[1][2] * im10
                + ureal[1][3] * re11
                - uimag[1][3] * im11
            )
            self.imag[ind01] = (
                uimag[1][0] * re00
                + ureal[1][0] * im00
                + uimag[1][1] * re01
                + ureal[1][1] * im01
                + uimag[1][2] * re10
                + ureal[1][2] * im10
                + uimag[1][3] * re11
                + ureal[1][3] * im11
            )

            self.real[ind10] = (
                ureal[2][0] * re00
                - uimag[2][0] * im00
                + ureal[2][1] * re01
                - uimag[2][1] * im01
                + ureal[2][2] * re10
                - uimag[2][2] * im10
                + ureal[2][3] * re11
                - uimag[2][3] * im11
            )
            self.imag[ind10] = (
                uimag[2][0] * re00
                + ureal[2][0] * im00
                + uimag[2][1] * re01
                + ureal[2][1] * im01
                + uimag[2][2] * re10
                + ureal[2][2] * im10
                + uimag[2][3] * re11
                + ureal[2][3] * im11
            )

            self.real[ind11] = (
                ureal[3][0] * re00
                - uimag[3][0] * im00
                + ureal[3][1] * re01
                - uimag[3][1] * im01
                + ureal[3][2] * re10
                - uimag[3][2] * im10
                + ureal[3][3] * re11
                - uimag[3][3] * im11
            )
            self.imag[ind11] = (
                uimag[3][0] * re00
                + ureal[3][0] * im00
                + uimag[3][1] * re01
                + ureal[3][1] * im01
                + uimag[3][2] * re10
                + ureal[3][2] * im10
                + uimag[3][3] * re11
                + ureal[3][3] * im11
            )

    def measure(self, target):
        zero_prob = self.calc_prob_of_outcome(target, 0)
        outcome, outcome_prob = self.generate_measure_outcome(zero_prob)
        self.collapse_to_know_prob_outcome(target, outcome, outcome_prob)
        return outcome

    def generate_measure_outcome(self, zero_prob):
        outcome = 0
        if zero_prob < REAL_EPS:
            outcome = 1
        elif (1 - zero_prob) < REAL_EPS:
            outcome = 0
        else:
            outcome = 1 if random.random() > zero_prob else 0

        outcome_prob = zero_prob if outcome == 0 else 1 - zero_prob
        return outcome, outcome_prob

    def collapse_to_know_prob_outcome(self, target, outcome, outcome_prob):
        num_task = self.total_num_amps // 2
        size_half_block = 2**target
        size_block = 2 * size_half_block

        renorm = 1 / math.sqrt(outcome_prob)
        if outcome == 0:
            for this_task in range(num_task):
                this_block = this_task // size_half_block
                index = this_block * size_block + this_task % size_half_block

                self.real[index] = self.real[index] * renorm
                self.imag[index] = self.imag[index] * renorm

                self.real[index + size_half_block] = 0
                self.imag[index + size_half_block] = 0
        else:
            for this_task in range(num_task):
                this_block = this_task // size_half_block
                index = this_block * size_block + this_task % size_half_block

                self.real[index] = 0
                self.imag[index] = 0

                self.real[index + size_half_block] = (
                    self.real[index + size_half_block] * renorm
                )
                self.imag[index + size_half_block] = (
                    self.imag[index + size_half_block] * renorm
                )

    def calc_prob_of_outcome(self, target, outcome):
        outcome_prob = self.find_prob_of_zero(target)
        if outcome == 1:
            outcome_prob = 1.0 - outcome_prob
        return outcome_prob

    def find_prob_of_zero(self, target):
        num_task = self.total_num_amps // 2
        size_half_block = 2**target
        size_block = size_half_block * 2
        total_prob = 0.0

        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index = this_block * size_block + this_task % size_half_block

            total_prob += (
                self.real[index] * self.real[index]
                + self.imag[index] * self.imag[index]
            )

        return total_prob

    def get_prob(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: index in state vector of probability amplitudes

        Returns:
            the probability of target index
        """
        if index < 0 or index >= self.total_num_amps:
            raise ValueError(f"{index} is illegal parameter.")

        real = self.real[index]
        imag = self.imag[index]
        # TODO:doing
        # print("in pysim=", real * real + imag * imag)
        return real * real + imag * imag

    def get_probs(self, qubits):
        """Get all probabilities of circuit.

        Returns:
            An array contains all probabilities of circuit.
        """
        num_outcome_probs = len(qubits)
        outcome_probs = [0] * (2**num_outcome_probs)
        for i in range(self.total_num_amps):
            outcome_ind = 0
            for q in range(num_outcome_probs):
                outcome_ind += self.extract_bit(qubits[q], i) * (2**q)

            real = self.real[i]
            imag = self.imag[i]
            prob = real * real + imag * imag
            outcome_probs[outcome_ind] += prob

        return outcome_probs

    # TODO: to matrix
    def get_statevector(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        # todo better in float or ndarray
        state_list = []
        for i in range(self.total_num_amps):
            real = self.real[i]
            imag = self.imag[i]
            if self.real[i] > -1e-15 and self.real[i] < 1e-15:
                real = 0
            if self.imag[i] > -1e-15 and self.imag[i] < 1e-15:
                imag = 0
            state = str(real) + ", " + str(imag)
            state_list.append(state)
        return state_list

    def apply_param_named_phase(
        self,
        qubits,
        num_qubits_per_reg,
        num_regs,
        encoding,
        phase_func_name,
        params,
        override_inds,
        override_phases,
        num_overrides,
        conj,
    ):
        i = 0
        max_num_regs_apply_arbitrary_phase = 100
        phase_inds = [0] * max_num_regs_apply_arbitrary_phase
        for index in range(self.total_num_amps):
            flat_ind = 0
            for r in range(num_regs):
                phase_inds[r] = 0
                if encoding == BitEncoding.UNSIGNED:
                    for q in range(num_qubits_per_reg[r]):
                        phase_inds[r] += (2**q) * self.extract_bit(
                            qubits[flat_ind], index
                        )
                        flat_ind += 1
                elif encoding == BitEncoding.TWOS_COMPLEMENT:
                    for q in range(num_qubits_per_reg[r] - 1):
                        phase_inds[r] += (2**q) * self.extract_bit(
                            qubits[flat_ind], index
                        )
                        flat_ind += 1
                    if self.extract_bit(qubits[flat_ind], index) == 1:
                        phase_inds[r] -= 2 ** (num_qubits_per_reg[r] - 1)
                    flat_ind += 1

            for i in range(num_overrides):
                found = 1
                for r in range(num_regs):
                    if phase_inds[r] != override_inds[i * num_regs + r]:
                        found = 0
                        break
                if 1 == found:
                    break

            phase = 0
            if i < num_overrides:
                phase = override_phases[i]
            else:
                if (
                    phase_func_name == PhaseFunc.NORM
                    or phase_func_name == PhaseFunc.INVERSE_NORM
                    or phase_func_name == PhaseFunc.SCALED_NORM
                    or phase_func_name == PhaseFunc.SCALED_INVERSE_NORM
                    or phase_func_name == PhaseFunc.SCALED_INVERSE_SHIFTED_NORM
                ):
                    norm = 0
                    if phase_func_name == PhaseFunc.SCALED_INVERSE_SHIFTED_NORM:
                        for r in range(num_regs):
                            norm += (phase_inds[r] - params[2 + r]) * (
                                phase_inds[r] - params[2 + r]
                            )
                    else:
                        for r in range(num_regs):
                            norm += phase_inds[r] * phase_inds[r]
                    norm = math.sqrt(norm)

                    if phase_func_name == PhaseFunc.NORM:
                        phase = norm
                    elif phase_func_name == PhaseFunc.INVERSE_NORM:
                        if norm == 0.0:
                            phase = params[0]
                        else:
                            phase = 1 // norm
                    elif phase_func_name == PhaseFunc.SCALED_NORM:
                        phase = params[0] * norm
                    elif (
                        phase_func_name == PhaseFunc.SCALED_INVERSE_NORM
                        or phase_func_name == PhaseFunc.SCALED_INVERSE_SHIFTED_NORM
                    ):
                        if norm <= REAL_EPS:
                            phase = params[1]
                        else:
                            phase = params[0] // norm
                elif (
                    phase_func_name == PhaseFunc.PRODUCT
                    or phase_func_name == PhaseFunc.INVERSE_PRODUCT
                    or phase_func_name == PhaseFunc.SCALED_PRODUCT
                    or phase_func_name == PhaseFunc.SCALED_INVERSE_PRODUCT
                ):
                    prod = 1
                    for r in range(num_regs):
                        prod *= phase_inds[r]

                    if phase_func_name == PhaseFunc.PRODUCT:
                        phase = prod
                    elif phase_func_name == PhaseFunc.INVERSE_PRODUCT:
                        if prod == 0.0:
                            phase = params[0]
                        else:
                            phase = 1 // prod
                    elif phase_func_name == PhaseFunc.SCALED_PRODUCT:
                        phase = params[0] * prod
                    elif phase_func_name == PhaseFunc.SCALED_INVERSE_PRODUCT:
                        if prod == 0.0:
                            phase = params[1]
                        else:
                            phase = params[0] // prod
                elif (
                    phase_func_name == PhaseFunc.DISTANCE
                    or phase_func_name == PhaseFunc.INVERSE_DISTANCE
                    or phase_func_name == PhaseFunc.SCALED_DISTANCE
                    or phase_func_name == PhaseFunc.SCALED_INVERSE_DISTANCE
                    or phase_func_name == PhaseFunc.SCALED_INVERSE_SHIFTED_DISTANCE
                ):
                    dist = 0
                    if phase_func_name == PhaseFunc.SCALED_INVERSE_SHIFTED_DISTANCE:
                        for r in range(0, num_regs, 2):
                            dist += (
                                phase_inds[r] - phase_inds[r + 1] - params[2 + r / 2]
                            ) * (phase_inds[r] - phase_inds[r + 1] - params[2 + r / 2])
                    else:
                        for r in range(0, num_regs, 2):
                            dist += (phase_inds[r + 1] - phase_inds[r]) * (
                                phase_inds[r + 1] - phase_inds[r]
                            )
                    dist = math.sqrt(dist)

                    if phase_func_name == PhaseFunc.DISTANCE:
                        phase = dist
                    elif phase_func_name == PhaseFunc.INVERSE_DISTANCE:
                        if dist == 0.0:
                            phase = params[0]
                        else:
                            phase = 1 // dist
                    elif phase_func_name == PhaseFunc.SCALED_DISTANCE:
                        phase = params[0] * dist
                    elif (
                        phase_func_name == PhaseFunc.SCALED_INVERSE_DISTANCE
                        or phase_func_name == PhaseFunc.SCALED_INVERSE_SHIFTED_DISTANCE
                    ):
                        if dist <= REAL_EPS:
                            phase = params[1]
                        else:
                            phase = params[0] // dist

            if conj:
                phase *= -1

            c = math.cos(phase)
            s = math.sin(phase)
            re = self.real[index]
            im = self.imag[index]

            self.real[index] = re * c - im * s
            self.imag[index] = re * s + im * c

    def insert_zero_bit(self, number, index):
        left = (number >> index) << index
        right = number - left
        return (left << 1) ^ right

    def insert_two_zero_bits(self, number, bit1, bit2):
        small = 0
        if bit1 < bit2:
            small = bit1
        else:
            small = bit2
        big = 0
        if bit1 < bit2:
            big = bit2
        else:
            big = bit1
        return self.insert_zero_bit(self.insert_zero_bit(number, small), big)

    def swap_qubit_amps(self, qb1, qb2):
        num_task = self.total_num_amps >> 2
        for this_task in range(num_task):
            ind00 = self.insert_two_zero_bits(this_task, qb1, qb2)
            ind01 = self.flip_bit(ind00, qb1)
            ind10 = self.flip_bit(ind00, qb2)
            re01 = self.real[ind01]
            im01 = self.imag[ind01]
            re10 = self.real[ind10]
            im10 = self.imag[ind10]
            self.real[ind01] = re10
            self.real[ind10] = re01
            self.imag[ind01] = im10
            self.imag[ind10] = im01

    def apply_qft(self, qubits, num_qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
            num_qubits: number of qubits
        """
        for q in range(num_qubits - 1, -1, -1):
            self.hadamard(qubits[q])

            if q == 0:
                break

            num_regs = 2
            num_qubits_per_reg = [q, 1]
            regs = [0] * 100
            for i in range(q + 1):
                regs[i] = qubits[i]

            params = [math.pi / (2**q)]

            self.apply_param_named_phase(
                regs,
                num_qubits_per_reg,
                num_regs,
                BitEncoding.UNSIGNED,
                PhaseFunc.SCALED_PRODUCT,
                params,
                None,
                None,
                0,
                0,
            )

        for i in range(num_qubits // 2):
            qb1 = qubits[i]
            qb2 = qubits[num_qubits - i - 1]
            self.swap_qubit_amps(qb1, qb2)

    def apply_full_qft(self):
        """
        Applies the quantum Fourier transform (QFT) to the entirety of qureg
        """
        qubits = [0] * 100
        for i in range(self.qubits):
            qubits[i] = i
        self.apply_qft(qubits, self.qubits)

    # TODO:need to improve.
    def paulix_local(self, work_real, work_imag, target_qubit):
        """pauli-X"""
        # jpaulix_local(target_qubit, self.total_num_amps, List(work_real), List(work_imag))
        size_half_block = 2**target_qubit
        size_block = 2 * size_half_block

        num_task = self.total_num_amps // 2
        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            state_real_up = work_real[index_up]
            state_imag_up = work_imag[index_up]

            work_real[index_up] = work_real[index_lo]
            work_imag[index_up] = work_imag[index_lo]

            work_real[index_lo] = state_real_up
            work_imag[index_lo] = state_imag_up

    def pauliy_local(self, work_real, work_imag, target_qubit, conj_fac):
        """pauli-Y"""
        size_half_block = 2**target_qubit
        size_block = 2 * size_half_block

        num_task = self.total_num_amps // 2
        for this_task in range(num_task):
            this_block = this_task // size_half_block
            index_up = this_block * size_block + this_task % size_half_block
            index_lo = index_up + size_half_block

            state_real_up = work_real[index_up]
            state_imag_up = work_imag[index_up]

            work_real[index_up] = conj_fac * work_imag[index_lo]
            work_imag[index_up] = conj_fac * (-work_real[index_lo])

            work_real[index_lo] = conj_fac * (-state_imag_up)
            work_imag[index_lo] = conj_fac * state_real_up

    def phase_shift_by_term(self, real, imag, target_qubit, term_real, term_imag):
        state_vec_size = self.total_num_amps
        cos_angle = term_real
        sin_angle = term_imag
        for index in range(state_vec_size):
            target_bit = self.extract_bit(target_qubit, index)
            if target_bit:
                state_real_lo = real[index]
                state_imag_lo = imag[index]

                real[index] = cos_angle * state_real_lo - sin_angle * state_imag_lo
                imag[index] = sin_angle * state_real_lo + cos_angle * state_imag_lo

    def pauliz_local(self, work_real, work_imag, target_qubit):
        term_real = -1
        term_imag = 0
        self.phase_shift_by_term(
            work_real, work_imag, target_qubit, term_real, term_imag
        )

    def calc_inner_product_local(self, bra_real, bra_imag, ket_real, ket_imag):
        inner_prod_real = 0
        inner_prod_imag = 0
        for index in range(self.total_num_amps):
            bra_re = bra_real[index]
            bra_im = bra_imag[index]
            ket_re = ket_real[index]
            ket_im = ket_imag[index]
            inner_prod_real += bra_re * ket_re + bra_im * ket_im
            inner_prod_imag += bra_re * ket_im - bra_im * ket_re
        return (inner_prod_real, inner_prod_imag)

    def get_expec_pauli_prod(self, pauli_prod_list):
        """
        Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: a list contains the indices of the target qubits,\
                the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            the expected value of a product of Pauli operators.
        """
        work_real = [0] * self.total_num_amps
        work_imag = [0] * self.total_num_amps
        for i in range(self.total_num_amps):
            work_real[i] = self.real[i]
            work_imag[i] = self.imag[i]
        for pauli_op in pauli_prod_list:
            op_type = pauli_op["oper_type"]
            if op_type == PauliOpType.PAULI_X.value:
                self.paulix_local(work_real, work_imag, pauli_op["target"])
            elif op_type == PauliOpType.PAULI_Y.value:
                self.pauliy_local(work_real, work_imag, pauli_op["target"], 1)
            elif op_type == PauliOpType.PAULI_Z.value:
                self.pauliz_local(work_real, work_imag, pauli_op["target"])

        real, imag = self.calc_inner_product_local(
            work_real, work_imag, self.real, self.imag
        )
        return real

    def get_expec_pauli_sum(self, oper_type_list, term_coeff_list):
        """
        Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: a list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: the coefficients of each term in the sum of Pauli products.

        Returns:
            the expected value of a sum of products of Pauli operators. 
        """
        """Computes the expected value of a sum of products of Pauli operators."""
        num_qb = self.qubits
        targs = []
        for q in range(num_qb):
            targs.append(q)

        value = 0
        idx = 0
        num_sum_terms = len(term_coeff_list)
        for t in range(num_sum_terms):
            pauli_prod_list = []
            for i in range(num_qb):
                temp = {}
                temp["oper_type"] = oper_type_list[idx]
                idx += 1
                temp["target"] = targs[i]
                pauli_prod_list.append(temp)
            value += term_coeff_list[t] * self.get_expec_pauli_prod(pauli_prod_list)

        return value



"""Test the dump function in QCircuit."""
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, CNOT


def test_dump_qusl():
    """Testing the functionality of serializing quantum circuits as QUSL files in JSON format."""
    qc = QCircuit()
    qreg = qc.allocate(2)

    H * qreg[0]
    CNOT * (qreg[0], qreg[1])
    Measure * qreg[0]
    Measure * qreg[1]

    qc.dump(file="a.qusl", format="qusl")
    expect_out = '"code": ["H * q[0]\\n", "MCX(1) * (q[0], q[1])\\n", "Measure * q[0]\\n", "Measure * q[1]\\n"]'

    with open("a.qusl", "r") as stream:
        container = stream.readline()
        circuit_out = container[container.rfind("code") - 1:-1]

    assert circuit_out == expect_out


def test_dump_openqasm():
    """Testing the functionality of serializing quantum circuits as OpenQASM files in JSON format."""
    qc = QCircuit()
    qreg = qc.allocate(2)

    H * qreg[0]
    CNOT * (qreg[0], qreg[1])
    Measure * qreg[0]
    Measure * qreg[1]

    qc.dump(file="b.qasm", format="openqasm")
    expect_out = ['OPENQASM 2.0;\n', 'include "qulib1.inc";\n', 'qreg q[2];\n', 'creg c[2];\n', 'h q[0];\n',
                  'cx q[0],q[1];\n', 'measure q[0] -> c[0];\n', 'measure q[1] -> c[1];\n']

    with open("b.qasm", "r") as stream:
        circuit_out = stream.readlines()

    assert circuit_out == expect_out



import cmath

import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class Z1Gate(BasicGate):
    """Apply the single-qubit Z1 gate.

    Example:
        .. code-block:: python

            Z1 * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Z1"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply Z1 gate.

        Example:
            .. code-block:: python

                Z1 * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [[np.exp(-1j * cmath.pi / 4), 0], [0, np.exp(1j * cmath.pi / 4)]]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = Z1Gate()
        gate.is_inverse = not self.is_inverse
        return gate


Z1 = Z1Gate()



"""Quantum Phase Estimation."""

from qutrunk.circuit.ops.operator import Operator
from qutrunk.circuit.gates import H, All, Barrier
from qutrunk.circuit.ops import IQFT


class QPE(Operator):
    """Quantum Phase Estimation.

    Args:
        unitary: Stand Gate.

    Example:
        .. code-block:: python

            from math import pi
            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import NOT, Barrier, P, All, Measure
            from qutrunk.circuit.ops import QPE

            def _bin_int(itrable):
                return int("".join(map(str, reversed(itrable))), base=2)

            # allocate
            qc = QCircuit()
            q1, q2 = qc.allocate([3, 1])

            # Prepare our eigenstate |psi>
            NOT * q2[0]
            Barrier * q1
            Barrier * q2

            # apply QPE
            QPE(P(pi/4)) * (q1, q2)

            # measure q1
            All(Measure) * q1

            # print circuit
            # qc.print()

            # run circuit
            qc.run(shots=100)

            # print result
            print(q1.to_cl())

            # calculate the value of theta
            f = _bin_int(q1.to_cl())
            theta = f / 2 ** len(q1)
            print("θ=", theta)

    """
    def __init__(self, unitary):
        """Initialize a QPE gate."""
        super().__init__()
        # TODO: whether to Gate.
        self.unitary = unitary

    def __str__(self):
        """Return a string representation of the object."""
        return f"QPE({str(self.unitary)})"

    def __mul__(self, qreg):
        qreg_first = qreg[0]
        qreg_second = qreg[1]

        # 1 apply H gate
        All(H) * qreg_first

        # 2 apply cu
        repetitions = 1
        num_qubits = len(qreg_first)
        for counting_qubit in range(num_qubits):
            for i in range(repetitions):
                # apply cu gate
                self.unitary.ctrl() * (qreg_first[counting_qubit], qreg_second[0])
            repetitions *= 2
        # 3 inverse QFT
        Barrier * qreg_first
        Barrier * qreg_second
        IQFT * qreg_first

        # 4 measure qreg_first
        # All(Measure) * qreg_first





#
# Autogenerated by Thrift Compiler (0.16.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
from .ttypes import *
ErrMsg = {
    4: "异常",
    2: "不合法参数",
    6: "存在",
    7: "缓存中",
    8: "内存不足",
    5: "不存在",
    9: "未初始化",
    10: "机器未注册",
    11: "无可用的资源",
    1: "未知错误",
    12: "数据解析异常",
    0: "成功",
    3: "超时",
}



import itertools

from .qubit import QuBit


class SubQureg:
    """SubQureg."""

    def __init__(self, circuit):
        self.circuit = circuit
        self.qubits = []

    def __getitem__(self, idx):
        """Return a Qubit instance."""
        if not isinstance(idx, int):
            raise ValueError("expected integer index into register")
        return self.qubits[idx]

    def append(self, qubit: QuBit):
        """Append QuBit to SubQureg.

        Args:
            qubit: The target qubit append to SubQureg.
        """
        self.qubits.append(qubit)

    def to_cl(self):
        """Get measure result of SubQureg"""
        start = self.qubits[0].index
        end = self.qubits[-1].index
        res = []
        for i in range(start, end + 1):
            res.append(self.circuit.creg[i].value)

        return res

    def __len__(self):
        return len(self.qubits)


class Qureg:
    """Register, maintains a set of qubits.

    Args:
        circuit: A quantum circuit.
        name: The name of the register.
        size: The size of the register.
    """

    prefix = "q"
    bit_type = "QuBit"
    instances_counter = itertools.count()

    def __init__(self, circuit=None, name=None, size=None):
        if name is None:
            name = f"{self.prefix}{self.instances_counter}"
        self.name = name

        self.qubits = []
        # TODO: delete size ?
        self.size = size
        self.circuit = circuit

        for index in range(self.size):
            self.qubits.append(QuBit(self, index))

    def __getitem__(self, idx):
        """Return a Qubit instance.

        Arg:
            idx: The index of Qubit.

        Returns:
            Qubit instance.

        Raises:
          ValueError: If the index of Qubit is not a integer data.
        """
        if not isinstance(idx, int):
            raise ValueError("expected integer index into register")
        return self.qubits[idx]

    def __len__(self):
        return len(self.qubits)

    def qasm(self):
        """Return OPENQASM string for this register."""
        return f"qreg q[{len(self)}];"

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.size})"

    def index(self, qubit):
        """Find the index of the provided qubit within this register.

        Args:
            qubit: The qubit to get index.
        """
        return self.qubits.index(qubit)

    def to_cl(self):
        """Get measure result of Qureg."""
        res = []
        for i in range(len(self.qubits)):
            res.append(self.circuit.creg[i].value)
        return res

    def split(self, sections: list):
        """Split Qureg into subqureg.

        Args:
            sections: Tell how to split the original Qureg, \
                each value item represents the size of corresponding subqureg.

        Returns:
            The tuple contains all subqureg.
        """
        if not isinstance(sections, list):
            raise TypeError("sections parameter should be a list.")

        if sum(sections) != len(self.qubits):
            raise ValueError("The sum of sections should be equal to len(qubits).")

        res = []
        start_index = 0
        for sec in sections:
            slices = SubQureg(self.circuit)
            for i in range(sec):
                # note: share the same QuBit
                slices.append(self.qubits[start_index + i])
            res.append(slices)
            start_index += sec

        return tuple(res)



"""Python implementation of a quantum computer simulator."""

from qutrunk.backends.backend import Backend
# TODO:need to improve.
from qutrunk.sim.local.local_python import BackendLocalPython as BackendLocalImpl


class BackendLocal(Backend):
    """Python implementation of a quantum computer simulator.

    The local backend uses the simulator to run the quantum circuit.

    Example:
        .. code-block:: python

            from qutrunk.backends import BackendLocal
            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import H, CNOT, Measure

            # new QCircuit object
            qc = QCircuit(backend=BackendLocal())
            # or use as default
            # qc = QCircuit()
            qr = qc.allocate(2)
            H * qr[0]
            CNOT * (qr[0], qr[1])
            Measure * qr[0]
            Measure * qr[1]
            res = qc.run(shots=100)
    """

    def __init__(self):
        super().__init__()
        self.circuit = None
        self._local_impl = BackendLocalImpl()

    def send_circuit(self, circuit, final=False):
        """Send the quantum circuit to local backend.

        Args:
            circuit: Quantum circuit to send.
            final: True if quantum circuit finish, default False, \
            when final==True The backend program will release the computing resources.
        """
        start = circuit.cmd_cursor
        stop = len(circuit.cmds)

        if start == 0:
            res, elapsed = self._local_impl.init(len(circuit.qreg))
            if self.circuit.counter:
                self.circuit.counter.acc_run_time(elapsed)

        res, elapsed = self._local_impl.send_circuit(circuit, final)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)

        circuit.forward(stop - start)

        return res

    def run(self, shots=1):
        """Run quantum circuit.

        Args:
            shots: Circuit run times, for sampling, default: 1.

        Returns:
            list: The Result object contain circuit running outcome.
        """
        res, elapsed = self._local_impl.run(shots)
        # TODO: circuit is None?
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
            self.circuit.counter.finish()
        return res

    def get_prob(self, index):
        """Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: Index in state vector of probability amplitudes.

        Returns:
            float:The probability of target index.
        """
        res, elapsed = self._local_impl.get_prob(index)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_probs(self, qubits):
        """Get all probabilities of circuit.

        Returns:
            list: An array contains all probabilities of circuit.
        """
        res, elapsed = self._local_impl.get_probs(qubits)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_statevector(self):
        """Get state vector of circuit.

        Returns:
            list: Array contains all amplitudes of state vector.
        """
        res, elapsed = self._local_impl.get_statevector()
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)

        return res

    def get_expec_pauli_prod(self, pauli_prod_list):
        """Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: A list contains the indices of the target qubits,\
                the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            float:The expected value of a product of Pauli operators.
        """
        res, elapsed = self._local_impl.get_expec_pauli_prod(pauli_prod_list)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_expec_pauli_sum(self, oper_type_list, term_coeff_list):
        """Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: A list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: The coefficients of each term in the sum of Pauli products.

        Returns:
            float:The expected value of a sum of products of Pauli operators.
        """
        res, elapsed = self._local_impl.get_expec_pauli_sum(
            oper_type_list, term_coeff_list
        )
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    @property
    def name(self):
        """The name of Backend."""
        return "BackendLocalPython"



from .basicgate import BasicGate
from qutrunk.circuit.qubit import QuBit
from qutrunk.circuit.command import Command


class BarrierGate(BasicGate):
    """Barrier Gate.

    Args:
        BasicGate: Base class of all gates.

    Example:
        .. code-block:: python

            Barrier * (qr[0], qr[1])
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Barrier"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: The quantum bit to apply Barrier gate.

        Example:
            .. code-block:: python

                Barrier * (qr[0], qr[1])
        """
        qubit = [qubits] if isinstance(qubits, QuBit) else qubits
        targets = [q.index for q in qubit]
        cmd = Command(self, targets)
        self.commit(qubit[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def label(self):
        """Return gate label."""
        return "Barrier"


Barrier = BarrierGate()



"""Swap gate."""

import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class SwapGate(BasicGate):
    """Performs a SWAP gate between qubit1 and qubit2.

    Example:
        .. code-block:: python

            Swap * (qr[0], qr[1])
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Swap"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: Qubits to swap.

        Example:
            .. code-block:: python

                Swap * (qr[0], qr[1])

        Raises:
            TypeError: If the argument is not a Qubit object.
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError(
                "Parameter error: Two target qubits are required."
            )

        targets = [q.index for q in qubits]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    def inv(self):
        """Apply inverse gate."""
        gate = SwapGate()
        gate.is_inverse = not self.is_inverse
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        if ctrl_cnt > 1:
            raise ValueError("Swap gate do not support multiple control bits.")
        gate = CSwapGate()
        gate.is_inverse = self.is_inverse
        return gate


Swap = SwapGate()


class CSwapGate(BasicGate):
    """Controlled-SWAP gate, also known as the Fredkin gate.

    Example:
        .. code-block:: python

            CSwap * (qr[0], qr[1], qr[2])
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "CSwap"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: qubits[0] is control qubit, qubits[1], qubits[2] is target qubits.

        Example:
            .. code-block:: python

                CSwap * (qr[0], qr[1], qr[2])
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 3:
            raise ValueError(
                "Parameter error: One controlled and two target qubits are required."
            )

        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index, qubits[2].index]
        cmd = Command(self, targets, controls, inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0, 1],
            ]
        )

    def inv(self):
        """Apply inverse gate"""
        gate = CSwapGate()
        gate.is_inverse = not self.is_inverse
        return gate


CSwap = CSwapGate()



"""Test the function of calculating the depth of quantum circuits."""
from numpy import pi

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, CNOT, iSwap, T, Swap, SqrtSwap, CH, All


def test_depth():
    """Test the depth function in QCircuit."""
    qc = QCircuit()
    qreg = qc.allocate(5)

    H * qreg[0]
    H * qreg[1]
    H * qreg[2]
    iSwap * (qreg[0], qreg[4])
    T * qreg[1]
    H * qreg[0]
    Swap * (qreg[1], qreg[4])
    CH * (qreg[0], qreg[1])
    SqrtSwap * (qreg[2], qreg[4])
    T * qreg[0]
    H * qreg[1]
    CNOT * (qreg[0], qreg[2])
    CH * (qreg[1], qreg[2])
    H * qreg[2]
    All(Measure) * qreg

    result = qc.depth()
    calculate = 9

    assert result == calculate








# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM CNOT statement."""

from .node import Node


class Cnot(Node):
    """Node for an OPENQASM CNOT statement.

    children[0], children[1] are id nodes if CX is inside a gate body,
    otherwise they are primary nodes.
    """

    def __init__(self, children):
        """Create the cnot node."""
        super().__init__("cnot", children, None)

    def qasm(self):
        """Return the corresponding OPENQASM string."""
        return "CX " + self.children[0].qasm() + "," + self.children[1].qasm() + ";"



"""Example of a simple quantum random number generator."""

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, All, X, CNOT


def deutsch(backend=None):
    # Create quantum circuit
    qc = QCircuit(backend=backend)

    # Allocate quantum qubits
    qureg = qc.allocate(2)

    # Apply quantum gates
    X * qureg[1]
    All(H) * qureg
    CNOT * (qureg[0], qureg[1])
    H * qureg[0]

    # Measure all quantum qubits
    All(Measure) * qureg

    # Print quantum circuit
    qc.print()

    # Run quantum circuit and print result
    result = qc.run()
    print("The result of Deutsch: ", result.get_measure()[0])

    return qc


if __name__ == "__main__":
    # Run locally
    circuit = deutsch()

    # Dram quantum circuit
    circuit.draw()



"""Quantum bit and Classical bit objects."""


class Bit:
    """A generic bit.

    Args:
        register: Bit's register.
        index: Bit's index.
    """

    def __init__(self, register=None, index=None):
        """Create a new generic bit."""
        if register is None and index is None:
            self.register = None
            self.index = None
        else:
            try:
                index = int(index)
            except Exception:
                raise TypeError(
                    f"index needs to be an int: type {type(index).__name__} was provided."
                )

            self.register = register
            self.index = index
            self._hash = hash((self.register, self.index))
            self._repr = f"{self.__class__.__name__}({self.register}, {self.index})"

    def __repr__(self):
        """Return the string representing the bit."""
        if self.register is None and self.index is None:
            return object.__repr__(self)

        return self._repr

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if self.register is None and self.index is None:
            return other is self

        try:
            return self._repr == other._repr
        except AttributeError:
            return False



import sys
from warnings import warn
import shutil

from qutrunk.circuit import CBit
from qutrunk.circuit.gates import (
    MeasureGate,
    HGate,
    XGate,
    BarrierGate,
    MCZ,
    MCX,
    P,
    U3,
    ZGate,
    R,
    Rx,
    Ry,
    Rz,
    SGate,
    SdgGate,
    TGate,
    TdgGate,
    YGate,
    SwapGate,
    SqrtXGate,
    CYGate,
    CRx,
    CRy,
    CRz,
    Rxx,
    Ryy,
    Rzz,
    U1,
    U2,
    U3,
    SqrtSwapGate,
    CP,
    X1Gate,
    Y1Gate,
    Z1Gate,
    CU,
    CU1,
    CU3,
    CR,
    iSwapGate,
    IGate,
    CHGate,
    CSwapGate,
    SqrtXdgGate,
    CSqrtXGate,
)
from .text_draw_element import InputWire
from .layer import Layer
from .text_draw_element import MeasureFrom, MeasureTo
from .text_draw_element import BoxOnQuWire
from .text_draw_element import Bullet, OpenBullet
from .text_draw_element import EmptyWire, BreakWire
from .text_draw_element import Barrier as BarrierBox
from .text_draw_element import Ex, iEx


class TextDrawing:
    """The text drawing.

    Args:
        qubits: Qubits.
        cbits: Classical bits.
        nodes: Operator nodes.
        circuit: Quantum circuit.
        plot_barriers: Wheter to draw barriers.
        line_length: Draw width.
        vertical_compression: Vertical compression type, available values: ["high", "medium", "low"].
        encoding: Encoding Type, default: UTF-8.

    Raises:
          ValueError.
    """

    def __init__(
        self,
        qubits,
        cbits,
        nodes,
        circuit=None,
        plot_barriers=True,
        line_length=None,
        vertical_compression="medium",
        encoding=None,
    ):
        # quantum qubits
        self.qubits = qubits
        # classics cbits
        self.cbits = cbits
        # operation node
        self.nodes = nodes
        # print barriers or not, default is true
        self.plot_barriers = plot_barriers
        # max line_length
        self.line_length = line_length
        # quantum circuit
        self.circuit = circuit

        # vertically compressed square matrix in text mode
        if vertical_compression not in ["high", "medium", "low"]:
            raise ValueError(
                "Vertical compression can only be 'high', 'medium', or 'low'"
            )
        self.vertical_compression = vertical_compression

        # encoding format
        if encoding:
            self.encoding = encoding
        else:
            if sys.stdout.encoding:
                self.encoding = sys.stdout.encoding
            else:
                self.encoding = "utf8"

    def __str__(self):
        return self.single_string()

    def __repr__(self):
        return self.single_string()

    def single_string(self):
        """Create a long string with the ascii art."""
        try:
            return "\n".join(self.lines()).encode(self.encoding).decode(self.encoding)
        except (UnicodeEncodeError, UnicodeDecodeError):
            warn(
                f"The encoding {self.encoding} has a limited charset. Consider a different encoding in your "
                "environment. UTF-8 is being used instead.",
                RuntimeWarning,
            )
            self.encoding = "utf-8"

            return "\n".join(self.lines()).encode(self.encoding).decode(self.encoding)

    def lines(self, line_length=None):
        """Generates a list with lines."""
        if line_length is None:
            line_length = self.line_length

        if not line_length:
            line_length, _ = shutil.get_terminal_size()

        # quantum num_qubits
        num_qubits = len(self.qubits)

        # TODO:build layers
        layers = self.build_layers()

        layer_groups = [[]]
        rest_of_the_line = line_length
        for layerno, layer in enumerate(layers):
            layers[layerno] = EmptyWire.fillup_layer(layer, num_qubits)

            TextDrawing.normalize_width(layer)

            if line_length == -1:
                layer_groups[-1].append(layer)
                continue

            # chop the layer to the line_length (pager)
            layer_length = layers[layerno][0].length

            if layer_length < rest_of_the_line:
                layer_groups[-1].append(layer)
                rest_of_the_line -= layer_length
            else:
                layer_groups[-1].append(BreakWire.fillup_layer(len(layer), "»"))

                # New group
                layer_groups.append([BreakWire.fillup_layer(len(layer), "«")])
                rest_of_the_line = line_length - layer_groups[-1][-1][0].length

                layer_groups[-1].append(
                    InputWire.fillup_layer(self.wire_names(with_initial_state=False))
                )
                rest_of_the_line -= layer_groups[-1][-1][0].length

                layer_groups[-1].append(layer)
                rest_of_the_line -= layer_groups[-1][-1][0].length

        lines = []
        for layer_group in layer_groups:
            wires = list(zip(*layer_group))
            lines += self.draw_wires(wires)

        # TODO:compression in layer
        # -| H |----------
        # --------| H |---
        # compression to:
        # -| H |---
        # -| H |---
        return lines

    # TODO:doing
    def node_to_gate(self, node, layer):
        """Turn the operation node into a corresponding gate operation \
            and connect the corresponding qubit bit.
        """
        # implement h, cx, measure gate
        op = node.op  # Command object
        current_cons = []
        connection_label = None
        # condition control
        conditional = False

        def add_connected_gate(node, gates, layer, current_cons):
            for i, gate in enumerate(gates):
                actual_index = self.qubits.index(node.qargs[i])
                if actual_index not in [i for i, j in current_cons]:
                    layer.set_qubit(node.qargs[i], gate)
                    current_cons.append((actual_index, gate))

        # Measure
        if isinstance(op.gate, MeasureGate):
            gate = MeasureFrom()
            layer.set_qubit(node.qargs[0], gate)
            register, _, reg_index = get_bit_reg_index(self.circuit, node.cargs[0])
            if register is not None:
                layer.set_cbit(
                    node.cargs[0],
                    MeasureTo(str(reg_index)),
                )
            else:
                layer.set_clbit(node.cargs[0], MeasureTo())
        # H gate
        elif isinstance(op.gate, HGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("H", conditional=conditional))
        # CH gate
        elif isinstance(op.gate, CHGate):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CH", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # I gate
        elif isinstance(op.gate, IGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("I", conditional=conditional))
        # CX gate
        elif isinstance(op.gate, MCX) and op.gate.ctrl_cnt == 1:
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CX", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # Toffoli gate
        elif isinstance(op.gate, MCX) and op.gate.ctrl_cnt == 2:
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("Toffoli", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # X gate
        elif isinstance(op.gate, XGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("X", conditional=conditional))
        # BarrierGate gate
        elif isinstance(op.gate, BarrierGate):
            for qubit in node.qargs:
                if qubit in self.qubits:
                    layer.set_qubit(qubit, BarrierBox())
        # CZ gate
        elif isinstance(op.gate, MCZ):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CZ", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # CX gate
        elif isinstance(op.gate, MCX):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("MCX", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        elif isinstance(op.gate, P):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("P", conditional=conditional))
        elif isinstance(op.gate, U3):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("U3", conditional=conditional))
        elif isinstance(op.gate, ZGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("Z", conditional=conditional))
        elif isinstance(op.gate, R):
            label = f"R({op.gate.theta:.2},{op.gate.phi:.2})"
            layer.set_qubit(node.qargs[0], BoxOnQuWire(label, conditional=conditional))
        elif isinstance(op.gate, Rx):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("Rx", conditional=conditional))
        elif isinstance(op.gate, Ry):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("Ry", conditional=conditional))
        elif isinstance(op.gate, Rz):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("Rz", conditional=conditional))
        elif isinstance(op.gate, SGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("S", conditional=conditional))
        elif isinstance(op.gate, SdgGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("S†", conditional=conditional))
        elif isinstance(op.gate, TGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("T", conditional=conditional))
        elif isinstance(op.gate, TdgGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("T†", conditional=conditional))
        elif isinstance(op.gate, YGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("Y", conditional=conditional))
        # 交互门Swap
        elif isinstance(op.gate, SwapGate):
            gates = [Ex(conditional=conditional) for _ in range(len(node.qargs))]
            add_connected_gate(node, gates, layer, current_cons)
        # 交互控制门CSwap
        elif isinstance(op.gate, CSwapGate):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.extend(
                [
                    Ex(conditional=conditional)
                    for _ in range(len(node.qargs) - len(gates))
                ]
            )
            add_connected_gate(node, gates, layer, current_cons)
        # SqrtX gate
        elif isinstance(op.gate, SqrtXGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("√X", conditional=conditional))
        # SqrtXdg gate
        elif isinstance(op.gate, SqrtXdgGate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("√X†", conditional=conditional))
        # CSqrtX gate
        elif isinstance(op.gate, CSqrtXGate):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("√X", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # CY gate
        elif isinstance(op.gate, CYGate):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CY", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # crx gate
        elif isinstance(op.gate, CRx):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("Rx(ϴ)", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # cry gate
        elif isinstance(op.gate, CRy):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("Ry(ϴ)", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # crz gate
        elif isinstance(op.gate, CRz):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("Rz(λ)", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # Rxx gate
        elif isinstance(op.gate, Rxx):
            gate_text = "Rxx(ϴ)"
            layer.set_qu_multibox(node.qargs, gate_text, conditional=conditional)
        # Ryy gate
        elif isinstance(op.gate, Ryy):
            gate_text = "Ryy(ϴ)"
            layer.set_qu_multibox(node.qargs, gate_text, conditional=conditional)
        # Rzz gate
        elif isinstance(op.gate, Rzz):
            gate_text = "Rzz(ϴ)"
            layer.set_qu_multibox(node.qargs, gate_text, conditional=conditional)
        elif isinstance(op.gate, U1):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("U1", conditional=conditional))
        elif isinstance(op.gate, U2):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("U2", conditional=conditional))
        elif isinstance(op.gate, U3):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("U3", conditional=conditional))
        # SqrtSwapGate
        elif isinstance(op.gate, SqrtSwapGate):
            gates = [Ex(conditional=conditional) for _ in range(len(node.qargs))]
            add_connected_gate(node, gates, layer, current_cons)
        # CP gate
        elif isinstance(op.gate, CP):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CP", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # X1 gate
        elif isinstance(op.gate, X1Gate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("X1", conditional=conditional))
        # Y1 gate
        elif isinstance(op.gate, Y1Gate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("Y1", conditional=conditional))
        # Z1 gate
        elif isinstance(op.gate, Z1Gate):
            layer.set_qubit(node.qargs[0], BoxOnQuWire("Z1", conditional=conditional))
        # CU gate
        elif isinstance(op.gate, CU):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CU", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # CU1 gate
        elif isinstance(op.gate, CU1):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CU1", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # CU3 gate
        elif isinstance(op.gate, CU3):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CU3", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # CR gate
        elif isinstance(op.gate, CR):
            ctrl_text = None
            params_array = TextDrawing.controlled_wires(node, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array

            gates = self.set_ctrl_state(
                node, conditional, ctrl_text, bool(controlled_bot)
            )
            gates.append(BoxOnQuWire("CR", conditional=conditional))
            add_connected_gate(node, gates, layer, current_cons)
        # iSwap
        elif isinstance(op.gate, iSwapGate):
            gates = [iEx(conditional=conditional) for _ in range(len(node.qargs))]
            add_connected_gate(node, gates, layer, current_cons)
        else:
            raise ValueError(
                "Text visualizer does not know how to handle this node: ", op.name
            )

        current_cons.sort(key=lambda tup: tup[0])
        current_cons = [g for q, g in current_cons]

        return layer, current_cons, connection_label

    def build_layers(self):
        """Constructs layers.

        Returns:
           list: List of DrawElements.
        """
        wire_names = self.wire_names()

        if not wire_names:
            return []

        layers = [InputWire.fillup_layer(wire_names)]

        # operator node
        for node_layer in self.nodes:
            # TODO:circuit
            layer = Layer(self.qubits, self.cbits, self.circuit)

            for node in node_layer:
                layer, current_connections, connection_label = self.node_to_gate(
                    node, layer
                )
                layer.connections.append((connection_label, current_connections))

            layer.connect_with("│")
            layers.append(layer.full_layer)

        return layers

    def _get_qubit_labels(self):
        """Qubit name format: q[index], eg: q[1]."""
        qubits = []
        for qubit in self.qubits:
            qubits.append(f"q[{qubit.index}]")
        return qubits

    def _get_cbit_labels(self):
        """Cbit name format: c[index], eg: c[1]."""
        cbits = []
        for cbit in self.cbits:
            cbits.append(f"c[{cbit.index}]")
        return cbits

    # TODO: optimization later
    def wire_names(self, with_initial_state=False):
        """Returns a list of names for each wire.

        Args:
            with_initial_state: If true, adds the initial value to the name.

        Returns:
            list:the list of wire name.
        """

        qubit_labels = self._get_qubit_labels()
        # TODO:doing
        # cbit_labels = self._get_cbit_labels()
        cbit_labels = "c: " + str(len(self.cbits)) + "/"

        if with_initial_state:
            qubit_labels = [f"{qubit}: |0>" for qubit in qubit_labels]
            # cbit_labels = [f"{cbit}: 0 " for cbit in cbit_labels]
        else:
            qubit_labels = [f"{qubit}: " for qubit in qubit_labels]
            # cbit_labels = [f"{cbit}: " for cbit in cbit_labels]

        return qubit_labels + [cbit_labels]

    def should_compress(self, top_line, bot_line):
        """Decides if the top_line and bot_line should be merged,\
            based on `self.vertical_compression`.
        """
        if self.vertical_compression == "high":
            return True

        if self.vertical_compression == "low":
            return False

        # self.vertical_compression == "medium":
        for top, bot in zip(top_line, bot_line):
            if top in ["┴", "╨"] and bot in ["┬", "╥"]:
                return False

        for line in (bot_line, top_line):
            no_spaces = line.replace(" ", "")
            if len(no_spaces) > 0 and all(
                c.isalpha() or c.isnumeric() for c in no_spaces
            ):
                return False

        return True

    @staticmethod
    def normalize_width(layer):
        """When the elements of the layer have different widths, sets the width to the max elements.

        Args:
            layer (list): A list of elements.
        """
        nodes = list(filter(lambda x: x is not None, layer))
        longest = max(node.length for node in nodes)
        for node in nodes:
            node.layer_width = longest

    def draw_wires(self, wires):
        """Creates a list of lines with the text drawing.

        Args:
            wires: [wire1, wire2, ...].

        Returns:
            list:[str1, str2, ...].
        """
        lines = []
        bottom_line = None
        for wire in wires:
            # 1 Build the top of the element
            top_line = ""
            for node in wire:
                top_line += node.top

            if bottom_line is None:
                lines.append(top_line)
            else:
                if self.should_compress(top_line, bottom_line):
                    lines.append(TextDrawing.merge_lines(lines.pop(), top_line))
                else:
                    lines.append(
                        TextDrawing.merge_lines(lines[-1], top_line, icod="bot")
                    )

            # 2 Build the middle of the element
            mid_line = ""
            for node in wire:
                # Build the middle of the element
                mid_line += node.mid
            lines.append(TextDrawing.merge_lines(lines[-1], mid_line, icod="bot"))

            # 3 Build the bottom of the element
            bottom_line = ""
            for node in wire:
                # Build the bottom of the element
                bottom_line += node.bot
            lines.append(TextDrawing.merge_lines(lines[-1], bottom_line, icod="bot"))

        return lines

    @staticmethod
    def merge_lines(top, bot, icod="top"):
        """Merges two lines."""
        ret = ""
        for topc, botc in zip(top, bot):
            if topc == botc:
                ret += topc
            elif topc in "┼╪" and botc == " ":
                ret += "│"
            elif topc == " ":
                ret += botc
            elif topc in "┬╥" and botc in " ║│" and icod == "top":
                ret += topc
            elif topc in "┬" and botc == " " and icod == "bot":
                ret += "│"
            elif topc in "╥" and botc == " " and icod == "bot":
                ret += "║"
            elif topc in "┬│" and botc == "═":
                ret += "╪"
            elif topc in "┬│" and botc == "─":
                ret += "┼"
            elif topc in "└┘║│░" and botc == " " and icod == "top":
                ret += topc
            elif topc in "─═" and botc == " " and icod == "top":
                ret += topc
            elif topc in "─═" and botc == " " and icod == "bot":
                ret += botc
            elif topc in "║╥" and botc in "═":
                ret += "╬"
            elif topc in "║╥" and botc in "─":
                ret += "╫"
            elif topc in "║╫╬" and botc in " ":
                ret += "║"
            elif topc == "└" and botc == "┌" and icod == "top":
                ret += "├"
            elif topc == "┘" and botc == "┐":
                ret += "┤"
            elif botc in "┐┌" and icod == "top":
                ret += "┬"
            elif topc in "┘└" and botc in "─" and icod == "top":
                ret += "┴"
            elif botc == " " and icod == "top":
                ret += topc
            else:
                ret += botc
        return ret

    @staticmethod
    def controlled_wires(node, layer):
        op = node.op
        # number of control qubit
        num_ctrl_qubits = len(op.controls)
        # The front num_ctrl_qubits is the control bit
        ctrl_qubits = node.qargs[:num_ctrl_qubits]
        # The post num_ctrl_qubits is the target bit
        args_qubits = node.qargs[num_ctrl_qubits:]
        ctrl_state = 2**num_ctrl_qubits - 1

        ctrl_state = f"{ctrl_state:b}".rjust(num_ctrl_qubits, "0")[::-1]

        in_box = []
        top_box = []
        bot_box = []

        qubit_index = sorted(i for i, x in enumerate(layer.qubits) if x in args_qubits)

        for ctrl_qubit in zip(ctrl_qubits, ctrl_state):
            if min(qubit_index) > layer.qubits.index(ctrl_qubit[0]):
                top_box.append(ctrl_qubit)
            elif max(qubit_index) < layer.qubits.index(ctrl_qubit[0]):
                bot_box.append(ctrl_qubit)
            else:
                in_box.append(ctrl_qubit)

        return (top_box, bot_box, in_box, args_qubits)

    def set_ctrl_state(self, node, conditional, ctrl_text, bottom):
        op = node.op
        gates = []
        # number of control qubit
        num_ctrl_qubits = len(op.controls)
        # The front num_ctrl_qubits is the control bit
        ctrl_qubits = node.qargs[:num_ctrl_qubits]
        ctrl_state = 2**num_ctrl_qubits - 1

        ctrl_state = f"{ctrl_state:b}".rjust(num_ctrl_qubits, "0")[::-1]

        for i in range(len(ctrl_qubits)):
            if ctrl_state[i] == "1":
                gates.append(
                    Bullet(conditional=conditional, label=ctrl_text, bottom=bottom)
                )
            else:
                gates.append(
                    OpenBullet(conditional=conditional, label=ctrl_text, bottom=bottom)
                )
        return gates


# TODO: have some problems
def get_wire_map(circuit, bits, cregbundle=True):
    """Map the bits and registers to the index.

    Args:
        circuit: The circuit being drawn.
        bits: The list of QuBit or CBit.
        cregbundle: If True bundle classical register. Default: True.

    Returns:
        Dict((QuBit, ClassicalRegister): index )
    """
    prev_reg = None
    wire_index = 0
    wire_map = {}

    for bit in bits:
        # index
        register = get_bit_register(circuit, bit)

        if register is None or not isinstance(bit, CBit) or not cregbundle:
            wire_map[bit] = wire_index
            wire_index += 1

        elif register is not None and cregbundle and register != prev_reg:
            prev_reg = register
            wire_map[register] = wire_index
            wire_index += 1
    #  dict((Qubit, ClassicalRegister): index)
    #  dict((Clbit, ClassicalRegister): index)
    return wire_map


def get_bit_register(circuit, bit):
    """Get the register for a bit.

    Args:
        circuit: The circuit being drawn.
        bit: Qubit or Cbit.

    Returns:
        Index(int)
    """
    return circuit.find_bit(bit)


def get_bit_reg_index(circuit, bit):
    """Get bit index and register information."""
    # index in quantum circuit diagrams
    bit_index = circuit.find_bit(bit)
    # Gets the register and the index in the register
    register = bit.register
    reg_index = bit.index

    return register, bit_index, reg_index



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class SqrtSwapGate(BasicGate):
    """Performs a sqrt SWAP gate between qubit1 and qubit2.

    Example:
        .. code-block:: python

            SqrtSwap * (qr[0], qr[1])
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "SqrtSwap"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: Qubits to swap.

        Example:
            .. code-block:: python

                SqrtSwap * (qr[0], qr[1])
        """
        targets = [q.index for q in qubits]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [1, 0, 0, 0],
                [0, 0.5 + 0.5j, 0.5 - 0.5j, 0],
                [0, 0.5 - 0.5j, 0.5 + 0.5j, 0],
                [0, 0, 0, 1],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = SqrtSwapGate()
        gate.is_inverse = not self.is_inverse
        return gate


SqrtSwap = SqrtSwapGate()



"""DAG represent the Quantum circuit."""
import itertools
from collections import OrderedDict

import retworkx as rx

from qutrunk.circuit import CBit, CReg, QuBit, Qureg
from .dagnode import DAGInNode, DAGOpNode, DAGOutNode


class DAGCircuit:
    """Convert quantum circuit to a directed acyclic graph(DAG)."""

    def __init__(self):
        """Create an empty circuit."""
        self.name = None
        self.metadata = None
        self._wires = set()
        self.input_map = OrderedDict()
        self.output_map = OrderedDict()
        self._multi_graph = rx.PyDAG()
        self.qubits = []
        self.cbits = []
        self._op_names = {}
        self.qregs = OrderedDict()
        self.cregs = OrderedDict()

    def _add_wire(self, wire):
        """Add a qubit or cbit to the circuit.

        Args:
            wire: The object of qubit or cbit.

        Raises:
            ValueError.
        """

        if wire not in self._wires:
            self._wires.add(wire)

            input_node = DAGInNode(wire=wire)
            output_node = DAGOutNode(wire=wire)
            input_map_id, output_map_id = self._multi_graph.add_nodes_from(
                [input_node, output_node]
            )
            input_node._node_id = input_map_id
            output_node._node_id = output_map_id

            self.input_map[wire] = input_node
            self.output_map[wire] = output_node
            self._multi_graph.add_edge(input_node._node_id, output_node._node_id, wire)
        else:
            raise ValueError(f"duplicate wire {wire}")

    def add_qubits(self, qubits):
        """Add qubits to DAGCircuit, and add wires.

        Args:
            qubits: The qubits add to DAG.

        Raises:
            ValueError.
        """
        if any(not isinstance(qubit, QuBit) for qubit in qubits):
            raise ValueError("not a QuBit instance.")

        duplicate_qubits = set(self.qubits).intersection(qubits)
        if duplicate_qubits:
            raise ValueError(f"duplicate qubits {duplicate_qubits}")

        self.qubits.extend(qubits)
        for qubit in qubits:
            self._add_wire(qubit)

    def add_cbits(self, cbits):
        """Add classical bits to DAGCircuit, and add wires.

        Args:
            cbits: The classical bits add to DAG.

        Raises:
            ValueError.
        """
        if any(not isinstance(cbit, CBit) for cbit in cbits):
            raise ValueError("not a CBit instance.")

        duplicate_cbits = set(self.cbits).intersection(cbits)
        if duplicate_cbits:
            raise ValueError(f"duplicate cbits {duplicate_cbits}")

        self.cbits.extend(cbits)
        for cbit in cbits:
            self._add_wire(cbit)

    def topological_nodes(self, key=None):
        """Return nodes in topological sort order.

        Args:
             key (Callable): A callable which will take a DAGNode object and
                 return a string sort key.

         Returns:
             Generator(DAGOpNode, DAGInNode, or DAGOutNode): node in topological order
        """

        def _key(x):
            return x.sort_key

        if key is None:
            key = _key

        return iter(rx.lexicographical_topological_sort(self._multi_graph, key=key))

    def topological_op_nodes(self, key=None):
        """Return op nodes in topological sort order.

        Args:
            key (Callable): A callable which will take a DAGNode object and
                return a string sort key.

        Returns:
            generator(DAGOpNode): Op node in topological order
        """
        return (nd for nd in self.topological_nodes(key) if isinstance(nd, DAGOpNode))

    @property
    def wires(self):
        """Return a list of the wires in order."""
        return self.qubits + self.cbits

    def _add_op_node(self, op, qargs, cargs):
        """Add operation node.

        Args:
            op: The operation associated with DAG node.
            qargs: A list of QuBit.
            cargs: A list of CBit.

        Returns:
            node_index: The integer node index for the new operation node on the DAG.
        """
        new_node = DAGOpNode(op=op, qargs=qargs, cargs=cargs)

        node_index = self._multi_graph.add_node(new_node)
        new_node._node_id = node_index
        self._increment_op(op)

        return node_index

    def apply_operation_back(self, op, qargs=None, cargs=None):
        """Add operation node at the end of DAG.

        Args:
            op (Commands): The operation associated with the DAG node.
            qargs (list[QuBit]): Qubits that op will be applied to.
            cargs (list[CBit]): Cbits that op will be applied to.

        Returns:
            DAGOpNode: The node for the op that was added to the dag.
        """
        if cargs is None:
            cargs = []

        if qargs is None:
            qargs = []

        node_index = self._add_op_node(op, qargs, cargs)

        all_bit = [qargs, cargs]
        self._multi_graph.insert_node_on_in_edges_multiple(
            node_index, [self.output_map[q]._node_id for q in itertools.chain(*all_bit)]
        )

        return self._multi_graph[node_index]

    def add_qreg(self, qreg):
        """Add all wires in a quantum register."""

        if not isinstance(qreg, Qureg):
            raise ValueError("not a Qureg instance.")

        if qreg.name in self.qregs:
            raise ValueError(f"duplicate register {qreg.name}")

        self.qregs[qreg.name] = qreg
        existing_qubits = set(self.qubits)
        for i in range(qreg.size):
            if qreg[i] not in existing_qubits:
                self.qubits.append(qreg[i])
                self._add_wire(qreg[i])

    def add_creg(self, creg):
        """Add all wires in a classical register."""
        if not isinstance(creg, CReg):
            raise ValueError("not a CReg instance.")

        if creg.name in self.cregs:
            raise ValueError(f"duplicate register {creg.name}")

        self.cregs[creg.name] = creg
        existing_cbits = set(self.cbits)
        for i in range(creg.size):
            if creg[i] not in existing_cbits:
                self.cbits.append(creg[i])
                self._add_wire(creg[i])

    def _increment_op(self, op):
        """Count the number of each door operation."""
        if op.name in self._op_names:
            self._op_names[op.name] += 1
        else:
            self._op_names[op.name] = 1



"""Addition operation."""

from qutrunk.circuit import Qureg
from qutrunk.circuit.ops.operator import Operator
from qutrunk.circuit.gates import X, MCX


class ADD(Operator):
    """Addition operation.

    Args:
        number: The value added to circuit.

    Example:
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import Measure, All
            from qutrunk.circuit.ops import Classical
            from qutrunk.circuit.ops import ADD

            circuit = QCircuit()
            qr = circuit.allocate(4)
            Classical(0) * qr
            ADD(3) * qr
            All(Measure) * qr
            res = circuit.run()
            print(res.get_outcome())
    """

    def __init__(self, number: int):
        super().__init__()
        if number < 0:
            raise ValueError("number must be greater than zero.")
        self.number = number

    def __mul__(self, qr: Qureg):
        if not isinstance(qr, Qureg):
            raise TypeError("The operand must be Qureg.")

        num_qubits = len(qr)


        for _ in range(self.number):
            ctrl = []
            for i in range(num_qubits, 1, -1):
                for j in range(i - 1):
                    ctrl.append(qr[j])
                MCX(i - 1) * (*ctrl, qr[i - 1])
                ctrl = []

            X * qr[0]



"""U3 gate."""

import numpy as np

from .basicgate import BasicGate
from .basicgate import BasicRotateGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class U3(BasicGate):
    """U3 gate.

    Args:
        theta: U3 gate parameter1.
        phi: U3 gate parameter2.
        lam: U3 gate parameter3.

    Example:
        .. code-block:: python

            U3(pi/2,pi/2,pi/2) * qr[0]
    """

    def __init__(self, theta, phi, lam):
        """Create new U3 gate.

        Args:
            theta: U3 gate parameter1.
            phi: U3 gate parameter2.
            lam: U3 gate parameter3.
        """
        super().__init__()

        if theta is None or phi is None or lam is None:
            raise ValueError("The argument cannot be empty.")
        self.theta = theta
        self.phi = phi
        self.lam = lam

    def __str__(self):
        return "U3"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply U3 gate.

        Example:
            .. code-block:: python

                U3(pi/2,pi/2,pi/2) * qr[0]

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(
            self,
            targets,
            rotation=[self.theta, self.phi, self.lam],
            inverse=self.is_inverse,
        )
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        theta = self.theta
        phi = self.phi
        lam = self.lam
        cos = np.cos(theta / 2)
        sin = np.sin(theta / 2)
        return np.array(
            [
                [cos, -np.exp(1j * lam) * sin],
                [np.exp(1j * phi) * sin, np.exp(1j * (phi + lam)) * cos],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = U3(self.theta, self.phi, self.lam)
        gate.is_inverse = not self.is_inverse
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        if ctrl_cnt > 1:
            raise ValueError("U3 gate do not support multiple control bits.")
        gate = CU3(self.theta, self.phi, self.lam)
        gate.is_inverse = self.is_inverse
        return gate


class CU3(BasicRotateGate):
    """Control U3 gate.

    Args:
            theta: U3 gate parameter 1.
            phi: U3 gate parameter 2.
            lam: U3 gate parameter 3.

    Example:
        .. code-block:: python

            CU3(pi/2,pi/2,pi/2) * (qr[0], qr[1])
    """

    def __init__(self, theta, phi, lam):
        """Create new CU3 gate.

        Args:
            theta: U3 gate parameter 1.
            phi: U3 gate parameter 2.
            lam: U3 gate parameter 3.
        """
        if theta is None or phi is None or lam is None:
            raise ValueError("The argument cannot be empty.")
        super().__init__()
        self.theta = theta
        self.phi = phi
        self.lam = lam

    def __str__(self):
        return "CU3"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply CU3 gate.

        Example:
            .. code-block:: python

                CU3(pi/2,pi/2,pi/2) * (qr[0], qr[1])

        Raises:
            TypeError: If the argument is not a Qubit object.
            ValueError: If the qubits should not be two.
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError("Parameter Error: One controlled and one target qubit are required.")

        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(
            self,
            targets,
            controls,
            inverse=self.is_inverse,
            rotation=[self.theta, self.phi, self.lam],
        )
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        # TODO: definition have problem.
        half_alpha = float(self.theta)
        half_beta = float(self.phi)
        half_theta = float(self.lam)
        cos = np.cos(half_alpha / 2)
        sin = np.sin(half_alpha / 2)
        return np.array(
            [
                [1, 0, 0, 0],
                [0, cos, 0, -np.exp(1j * half_theta) * sin],
                [0, 0, 1, 0],
                [
                    0,
                    np.exp(1j * half_beta) * sin,
                    0,
                    np.exp(1j * (half_beta + half_theta)) * cos,
                ],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = CU3(self.theta, self.phi,  self.lam)
        gate.is_inverse = not self.is_inverse
        return gate


class CU(BasicRotateGate):
    """Control U gate.

    Args:
        theta: U gate parameter 1.
        phi: U gate parameter 2.
        lam:U gate parameter 3.
        gamma: U gate parameter 4.

    Example:
        .. code-block:: python

            CU(pi/2,pi/2,pi/2,pi/2) * (qr[0], qr[1])
    """

    def __init__(self, theta, phi, lam, gamma):
        """
        Args:
            theta: U gate parameter 1.
            phi: U gate parameter 2.
            lam:U gate parameter 3.
            gamma: U gate parameter 4.
        """
        if theta is None or phi is None or lam is None or gamma is None:
            raise ValueError("The argument cannot be empty.")
        super().__init__()
        self.theta = theta
        self.phi = phi
        self.lam = lam
        self.gamma = gamma

    def __str__(self):
        return "CU"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply CU gate.

        Example:
            .. code-block:: python

                CU(pi/2,pi/2,pi/2,pi/2) * (qr[0], qr[1])
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError("Parameter error: One controlled and one target qubit is required.")

        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(
            self,
            targets,
            controls,
            inverse=self.is_inverse,
            rotation=[self.theta, self.phi, self.lam, self.gamma],
        )
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        half_theta = float(self.theta)
        half_phi = float(self.phi)
        half_lam = float(self.lam)
        half_gamma = float(self.gamma)
        cos = np.cos(half_theta / 2)
        sin = np.sin(half_theta / 2)
        a = np.exp(1j * half_gamma) * cos
        b = -np.exp(1j * (half_gamma + half_lam)) * sin
        c = np.exp(1j * (half_gamma + half_phi)) * sin
        d = np.exp(1j * (half_gamma + half_phi + half_lam)) * cos
        return np.array(
            [
                [1, 0, 0, 0],
                [0, a, 0, b],
                [0, 0, 1, 0],
                [0, c, 0, d],
            ]
        )

    def inv(self):
        """Apply inverse gate"""
        gate = CU(self.theta, self.phi, self.lam, self.gamma)
        gate.is_inverse = not self.is_inverse
        return gate



"""The S gate."""

import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class SGate(BasicGate):
    """Apply the single-qubit S gate.

    Example:
        .. code-block:: python

            S * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "S"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply S gate.

        Example:
            .. code-block:: python

                S * qr[0]

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0], [0, 1j]])

    def inv(self):
        """Apply inverse gate."""
        gate = SGate()
        gate.is_inverse = not self.is_inverse
        return gate


S = SGate()



import numpy as np

from .basicgate import BasicRotateGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class CR(BasicRotateGate):
    """Control rotation gate.

    Args:
        alpha: Rotation angle.

    Example:
        .. code-block:: python

            CR(pi/2) * (qr[0], qr[1])
    """

    def __init__(self, alpha):
        """
        Args:
            alpha: Rotation angle.
        """
        if alpha is None:
            raise TypeError("The argument cannot be empty.")
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "CR"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply CR gate.

        Example:
            .. code-block:: python

                CR(pi/2) * (qr[0], qr[1])

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
            AttributeError: If the qubits should not be two.
        """         
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError("Parameter error: One controlled and one target qubit are required.")

        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(
            self, targets, controls, inverse=self.is_inverse, rotation=[self.rotation]
        )
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        half_alpha = float(self.rotation)
        return np.matrix(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, np.exp(1j * half_alpha)],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        # note: 状态相关方法需要重新生成一个新对象
        gate = CR(self.rotation)
        gate.is_inverse = not self.is_inverse 
        return gate


"""Get Qubox configure info."""

import os
from pathlib import Path

import yaml


def get_qubox_setting():
    """Get Qubox configure info."""
    cur_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    yaml_file = Path(cur_path) / "config" / "qubox.yaml"

    with open(yaml_file) as f:
        yaml_content = yaml.load(f, Loader=yaml.FullLoader)
        return yaml_content



"""Testing the functionality of parsing OpenQASM quantum circuits into QuTrunk quantum circuits."""
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, CNOT


def test_qasm_file():
    """Test the from_qasm_file function in QCircuit."""
    qc = QCircuit()
    qreg = qc.allocate(2)

    H * qreg[0]
    CNOT * (qreg[0], qreg[1])
    Measure * qreg[0]
    Measure * qreg[1]

    expect_in = qc.cmds

    result = qc.load(file='b.qasm', format="openqasm")
    circuit_in = result.cmds

    assert expect_in == circuit_in




import json
import sys
from subprocess import Popen, PIPE, TimeoutExpired


def read_ide_parameter(file_name: str):
    """Read the JSON file provided by QuIDE.

    Args:
        file_name: JSON file.

    Returns:
        file: Python dictionary.
    """
    # TODO:读取保内数据文件 pkgutil.get_data()
    with open(file=file_name, encoding="utf-8") as f:
        file = json.load(f)
    return file


def algorithm(file: dict):
    """According to IDE_Parameter.json generate quantum circuit.

    Args:
        file: Json file tell how to make quantum circuit.

    Returns:
        File path to store the quantum circuit.
    """
    with open(file="algorithm.py", mode="w", encoding="utf-8") as f:
        f.write("import json\n")
        f.write("from collections import defaultdict\n")
        f.write("from qutrunk.circuit import QCircuit\n")
        gt = "from qutrunk.circuit.gates import " + ",".join(file["gates"])
        f.write(gt + "\n")
        f.write("qc = QCircuit()\n")

        f.write(f"qr = qc.allocate({file['qubit']})\n")
        for c in file["cmd"]:
            f.write(c + "\n")

        f.write("outcome = qc.get_probs()\n")
        # f.write("print(qc.get_probs())\n")
        f.write("qc.run()\n")
        # f.write("print([int(q) for q in qr])\n")

        f.write("Tree = lambda: defaultdict(Tree)\n")
        f.write("tree = Tree()\n")
        c = [
            "for _, item in enumerate(outcome):\n",
            "    i = str(bin(item['idx']))[2:]\n",
            "    tree['probs'][i]['probability'] = item['prob']\n",
            "    tree['probs'][i]['angle'] = 0\n",
        ]
        f.writelines(c)
        # TODO:后期删除
        f.write("print(json.dumps(tree))\n")
        result_path = file["result_path"]
        write_file = [
            f"with open(file=r'{result_path}', mode='w', encoding='utf-8') as f:\n",
            "    file = json.dump(json.dumps(tree), f)\n",
        ]
        f.writelines(write_file)


def run_algorithm(params_file: str):
    """Run Python scripts using Tornado."""
    # 使用Tornado运行Python脚本
    # 1 读取IDE提供的参数JSON文件
    file = read_ide_parameter(params_file)

    # 2 根据ide_parameter.json生成量子算法
    algorithm(file)

    # 3 调用量子算法Python脚本
    proc = Popen([sys.executable, "algorithm.py"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
    try:
        outs, errs = proc.communicate(timeout=15)
    except TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()

    if errs:
        print("run error in algorithm.py")
        return
    print(outs.decode())
    print("sucess")


if __name__ == "__main__":
    run_algorithm("ide_parameter.json")



from enum import Enum

from qutrunk.backends.backend import Backend
from qutrunk.tools.read_qubox import get_qubox_setting
from .rpcclient import QuSproutApiServer
from qutrunk.sim.qusprout.qusproutdata import ttypes as qusproutdata


class ExecType(Enum):
    """Init exec type for quantum circuit.

    Args:
        ExecBySingleProcess: Execute by single process.
        ExecByMpi: Execute by multiple processes.
    """

    SingleProcess = 1
    Mpi = 2


class BackendQuSprout(Backend):
    """
    QuSprout: quamtum circuit simulator, provide multi-threaded OMP, multi node parallel MPI, GPU hardware acceleration.
    To use qusprout, make sure the network is connected and the service IP and Port are set correctly.

    Args:
        exectype: SingleProcess: use single calculation node; Mpi: parallel calculation using multiple nodes.

    Example:
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.backends import BackendQuSprout
            from qutrunk.circuit.gates import H, CNOT, Measure

            # use BackendQuSprout
            qc = QCircuit(backend=BackendQuSprout())
            qr = qc.allocate(2)

            # apply gate
            H * qr[0]
            CNOT * (qr[0], qr[1])
            Measure * qr[0]
            Measure * qr[1]

            # run circuit
            res = qc.run(shots=100)

            # print result
            print(res.get_counts())
    """

    def __init__(self, exectype=ExecType.SingleProcess):
        super().__init__()
        self.circuit = None
        self.exectype = exectype
        box_config = get_qubox_setting()
        self._api_server = QuSproutApiServer(
            ip=box_config.get("ip"), port=box_config.get("port")
        )

    def get_prob(self, index):
        """Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: Index in state vector of probability amplitudes.

        Returns:
            The probability of target index.
        """
        res, elapsed = self._api_server.get_prob(index)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_probs(self, qubits):
        """Get all probabilities of circuit.

        Returns:
            An array contains all probabilities of circuit.
        """
        res, elapsed = self._api_server.get_probs(qubits)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_statevector(self):
        """Get the current state vector of probability amplitudes for a set of qubits.

        Returns:
            Array contains all amplitudes of state vector
        """
        res, elapsed = self._api_server.get_statevector()
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def send_circuit(self, circuit, final=False):
        """Send the quantum circuit to qusprout backend.

        Args:
            circuit: Quantum circuit to send.
            final: True if quantum circuit finish, default False, \
            when final==True The backend program will release the computing resources.
        """
        cmds = []
        start = circuit.cmd_cursor
        stop = len(circuit.cmds)

        for idx in range(start, stop):
            cmd = circuit.cmds[idx]

            cmdex = None
            if cmd.cmdex is not None:
                _amp = None
                _mat = None
                if cmd.cmdex.amp is not None:
                    _amp = qusproutdata.Amplitude(
                        cmd.cmdex.amp.reals,
                        cmd.cmdex.amp.imags,
                        cmd.cmdex.amp.startind,
                        cmd.cmdex.amp.numamps,
                    )
                if cmd.cmdex.mat is not None:
                    _mat = qusproutdata.Matrix(
                        cmd.cmdex.mat.reals, cmd.cmdex.mat.imags, cmd.cmdex.mat.unitary
                    )
                cmdex = qusproutdata.Cmdex(amp=_amp, mat=_mat)

            c = qusproutdata.Cmd(
                str(cmd.gate),
                cmd.targets,
                cmd.controls,
                cmd.rotation,
                cmd.qasm(),
                cmd.inverse,
                cmdex,
            )
            cmds.append(c)

        circuit.forward(stop - start)

        # 服务端初始化
        if start == 0:
            res, elapsed = self._api_server.init(
                circuit.num_qubits, circuit.density, self.exectype.value
            )
            if self.circuit.counter:
                self.circuit.counter.acc_run_time(elapsed)

        if len(cmds) == 0 and (not final):
            return

        # 发送至服务
        res, elapsed = self._api_server.send_circuit(qusproutdata.Circuit(cmds), final)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)

    def run(self, shots=1):
        """Run quantum circuit.

        Args:
            shots: Circuit run times, for sampling, default: 1.

        Returns:
            The Result object contain circuit running outcome.
        """
        res, elapsed = self._api_server.run(shots)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
            self.circuit.counter.finish()

        """
        1 必须释放连接，不然其它连接无法连上服务端
        2 不能放在__del__中，因为对象释放不代表析构函数会及时调用
        """
        self._api_server.close()

        return res

    def get_expec_pauli_prod(self, pauli_prod_list):
        """Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: A list contains the indices of the target qubits,\
            the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            The expected value of a product of Pauli operators.
        """
        puali_list = []
        for temp in pauli_prod_list:
            puali_list.append(
                qusproutdata.PauliProdInfo(temp["oper_type"], temp["target"])
            )

        res, elapsed = self._api_server.get_expec_pauli_prod(puali_list)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_expec_pauli_sum(self, oper_type_list, term_coeff_list):
        """Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: A list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: The coefficients of each term in the sum of Pauli products.

        Returns:
            The expected value of a sum of products of Pauli operators.
        """
        res, elapsed = self._api_server.get_expec_pauli_sum(
            oper_type_list, term_coeff_list
        )
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    @property
    def name(self):
        return "BackendQuSprout"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM measure statement."""
from .node import Node


class Measure(Node):
    """Node for an OPENQASM measure statement.

    children[0] is a primary node (id or indexedid)
    children[1] is a primary node (id or indexedid)
    """

    def __init__(self, children):
        """Create the measure node."""
        super().__init__("measure", children, None)

    def qasm(self):
        """Return the corresponding OPENQASM string."""
        return (
            "measure "
            + self.children[0].qasm()
            + " -> "
            + self.children[1].qasm()
            + ";"
        )



from .bit import Bit


class QuBit(Bit):
    """A quantum bit.

    Args:
        qreg(Qureg): A quantum register containing the bit.
        index (int): The index of the bit in its containing register.

    Raises:
        TypeError: If the provided register is not a Qureg object.
    """

    def __init__(self, qreg=None, index=None):
        """Creates a quantum bit."""
        from .qureg import Qureg

        if not isinstance(qreg, Qureg):
            raise TypeError(
                f"QuBit needs a Qureg and {type(qreg).__name__} was provided."
            )

        super().__init__(qreg, index)
        if qreg is not None:
            self.circuit = qreg.circuit



"""QuSL parse."""

import os
import json
import random
import importlib

from qutrunk.exceptions import QuTrunkError


def _parse(file):
    """Parse QuSL file and generate quantum circuit.

    Args:
        file: The input QuSL file(json format).

    Returns:
        QCircuit object.
    """

    with open(file, mode="r", encoding="utf-8") as f:
        qusl_data = json.load(f)

    if qusl_data is None:
        raise ValueError("Empty QuSL file.")

    qubits = qusl_data["meta"]["qubits"]
    qusl_code_lines = qusl_data["code"]

    if int(qubits) <= 0:
        raise ValueError("Invalid qubits.")

    name = random.randint(1, 10000)
    filename = f"{name}.py"

    with open(file=filename, mode="w", encoding="utf-8") as fw:
        fw.write("from qutrunk.circuit import QCircuit\n")
        fw.write("from qutrunk.circuit.gates import *\n")
        fw.write("def generate_circuit():\n")
        fw.write("\tcircuit = QCircuit()\n")
        fw.write(f"\tq = circuit.allocate({qubits})\n")
        for line in qusl_code_lines:
            fw.write("\t" + line)
        fw.write("\treturn circuit\n\n")

    return str(name)


def qusl_to_circuit(file):
    """Parse QuSL file and generate quantum circuit.

    Args:

        file: The input QuSL file(json format).

    Returns:
        QCircuit object.
    """

    try:
        # import generate_circuit function.
        filename = _parse(file)
        g = importlib.import_module(filename)
        # generate circuit object.
        circuit = g.generate_circuit()
        # delete the temp file
        os.remove(f"{filename}.py")
    except Exception as e:
        raise QuTrunkError(e)
    else:
        return circuit




import cmath
from typing import Optional

from qutrunk.circuit.command import Command, CmdEx, Amplitude
from qutrunk.circuit import Qureg
from qutrunk.circuit.ops import Operator 


class AMP(Operator):
    """Quantum state preparation Operator.

    Init the quantum state to specific amplitude state.

    Args:
        classicvector: The amplitude state list.
        startind: The amplitude start index.
        numamps: The number of amplitude.

    Example:
        .. code-block:: python

            from qutrunk.circuit.ops import AMP
            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import H, All, Measure

            circuit = QCircuit()
            qureg = circuit.allocate(2)
            AMP([1-2j, 2+3j, 3-4j, 0.5+0.7j], 1, 2) * qureg
            print(circuit.get_statevector())
    """

    def __init__(
        self,
        classicvector: list,
        startind: Optional[int] = None,
        numamps: Optional[int] = None,
    ):
        super().__init__()
        self.classicvector = classicvector
        self.startind = startind
        self.numamps = numamps

    def __str__(self):
        return "AMP"

    def __mul__(self, qureg: Qureg):
        """Apply the AMP encode operator."""
        if not isinstance(qureg, Qureg):
            raise TypeError("the operand must be Qureg.")

        if not self._check_state(qureg):
            raise ValueError(f"Invalid state: {self.classicvector}, {self.startind}, {self.numamps}")

        self._process_state(qureg)

    def _check_state(self, qureg: Qureg):
        if self.startind is None:
            self.startind = 0

        if self.numamps is None or self.numamps > len(self.classicvector):
            self.numamps = len(self.classicvector)

        if (
            0 <= len(self.classicvector) <= 2 ** len(qureg)
            and 0 <= self.startind < 2 ** len(qureg)
            and 0 <= self.numamps <= 2 ** len(qureg)
            and (self.startind + self.numamps) <= 2 ** len(qureg)
        ):
            return True

        return False

    def _process_state(self, qureg: Qureg):
        """Process amp state."""
        reals = []
        imags = []
        listsum = sum(self.classicvector)
        for element in self.classicvector:
            normalized_element = cmath.sqrt(complex(element / listsum))
            reals.append(normalized_element.real)
            imags.append(normalized_element.imag)

        cmd = Command(self, cmdex=CmdEx(amp=Amplitude()))
        cmd.cmdex.amp.reals = reals
        cmd.cmdex.amp.imags = imags
        cmd.cmdex.amp.startind = self.startind
        cmd.cmdex.amp.numamps = self.numamps

        self.commit(qureg.circuit, cmd)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM external function."""

import numpy as np

from .node import Node
from .nodeexception import NodeException


class External(Node):
    """Node for an OPENQASM external function.

    children[0] is an id node with the name of the function.
    children[1] is an expression node.
    """

    def __init__(self, children):
        """Create the external node."""
        super().__init__("external", children, None)

    def qasm(self):
        """Return the corresponding OPENQASM string."""
        return self.children[0].qasm() + "(" + self.children[1].qasm() + ")"

    def latex(self):
        """Return the corresponding math mode latex string."""
        try:
            from pylatexenc.latexencode import utf8tolatex
        except ImportError as ex:
            raise Exception(
                "pylatexenc", "latex-from-qasm exporter", "pip install pylatexenc"
            ) from ex
        return utf8tolatex(self.sym())

    def real(self, nested_scope=None):
        """Return the correspond floating point number."""
        op = self.children[0].name
        expr = self.children[1]
        dispatch = {
            "sin": np.sin,
            "cos": np.cos,
            "tan": np.tan,
            "asin": np.arcsin,
            "acos": np.arccos,
            "atan": np.arctan,
            "exp": np.exp,
            "ln": np.log,
            "sqrt": np.sqrt,
        }
        if op in dispatch:
            arg = expr.real(nested_scope)
            return dispatch[op](arg)
        else:
            raise NodeException("internal error: undefined external")

    def sym(self, nested_scope=None):
        """Return the corresponding symbolic expression."""
        op = self.children[0].name
        expr = self.children[1]
        dispatch = {
            "sin": np.sin,
            "cos": np.cos,
            "tan": np.tan,
            "asin": np.arcsin,
            "acos": np.arccos,
            "atan": np.arctan,
            "exp": np.exp,
            "ln": np.log,
            "sqrt": np.sqrt,
        }
        if op in dispatch:
            arg = expr.sym(nested_scope)
            return dispatch[op](arg)
        else:
            raise NodeException("internal error: undefined external")



from math import pi
from typing import Iterable, Union

from qutrunk.circuit import QuBit, Qureg
from qutrunk.circuit.gates import CP, H, Swap
from qutrunk.circuit.ops.operator import Operator


class QFTOps(Operator):
    """Quantum Fourier Transfer Operator.

    Example:
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import All, Measure
            from qutrunk.circuit.ops import QFT

            circuit = QCircuit()
            qreg = circuit.allocate(5)

            # for full qft
            # QFT * qreg
            qubits = list(qreg)[::-2]
            QFT * qubits

            print(circuit.draw(line_length=1000))
            state = circuit.get_statevector()
            print(state)

            All(Measure) * qreg

            res = circuit.run(shots=1000)
            print(res.get_counts())

    """

    def __init__(self):
        super().__init__()

    def __mul__(self, qubits: Union[Qureg, Iterable[QuBit]]):
        """
        Args:
            qubits: Union[Qureg, Iterable[QuBit]], can be partial QFT.
        """
        if not all(isinstance(qb, QuBit) for qb in qubits):
            raise TypeError("The operand must be Qureg or Iterable of QuBit.")

        qb_cnt = len(qubits)
        for ctrl_qb_num in reversed(range(qb_cnt)):
            H * qubits[ctrl_qb_num]

            for target_qb_num in reversed(range(ctrl_qb_num)):
                # Use negative exponents so that the angle safely underflows to zero, rather than
                # using a temporary variable that overflows to infinity in the worst case.
                lam = pi * (2.0 ** (target_qb_num - ctrl_qb_num))
                CP(lam) * (qubits[ctrl_qb_num], qubits[target_qb_num])
        for i in range(qb_cnt // 2):
            Swap * (qubits[i], qubits[qb_cnt - i - 1])


class IQFTOps(Operator):
    """Inverse Quantum Fourier Transfer Operator.

    Example:
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import All, Measure
            from qutrunk.circuit.ops import IQFT

            circuit = QCircuit()
            qreg = circuit.allocate(5)

            IQFT * qreg

            print(circuit.draw(line_length=1000))
            state = circuit.get_statevector()
            print(state)

            All(Measure) * qreg

            res = circuit.run(shots=1000)
            print(res.get_counts())

    """
    def __init__(self):
        super().__init__()

    def __mul__(self, qubits: Union[Qureg, Iterable[QuBit]]):
        """
        Args:
            qubits: Union[Qureg, Iterable[QuBit]], can be partial QFT.
        """
        qb_cnt = len(qubits)
        for i in range(qb_cnt // 2):
            # swap is same as its inverse
            Swap * (qubits[i], qubits[qb_cnt - i - 1])

        for target_qb_num in range(qb_cnt):
            # H is same as its inverse
            H * qubits[target_qb_num]

            for ctrl_qb_num in range(target_qb_num + 1, qb_cnt):
                lam = -pi * (2.0 ** (target_qb_num - ctrl_qb_num))
                CP(lam) * (qubits[ctrl_qb_num], qubits[target_qb_num])


QFT = QFTOps()
IQFT = IQFTOps()



"""Classical Bit reference object."""
from .bit import Bit


class CBit(Bit):
    """A classical bit.

    Args:
        creg(CReg): A classical register containing the bit.
        index (int): The index of the bit in its containing register.
        value: The value of qubit which is measured.

    Raises:
        TypeError: If the provided register is not a CReg object.
    """

    def __init__(self, creg=None, index=None, value=None):
        """Creates a classical bit."""
        from .classical_reg import CReg

        if not isinstance(creg, CReg):
            raise TypeError(
                f"CBit needs a CReg and {type(creg).__name__} was provided."
            )

        super().__init__(creg, index)
        self._value = value

        if creg is not None:
            self.circuit = creg.circuit

    @property
    def value(self):
        """Get cbit value."""
        return self._value

    @value.setter
    def value(self, value):
        """Update cbit value when the corresponding qubit measure done."""
        self._value = value



from qutrunk.circuit.gates import All, H
from qutrunk.circuit import Qureg
from qutrunk.circuit.ops import Operator
from qutrunk.exceptions import QuTrunkError


class Plus(Operator):
    """Quantum state preparation Operator.

    Init the quantum state to plus state.

    Example:
        .. code-block:: python

            from qutrunk.circuit.ops import PLUS
            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import H, All, Measure

            circuit = QCircuit()
            qureg = circuit.allocate(2)
            PLUS * qureg
            print(circuit.get_statevector())
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "PLUS"

    def __mul__(self, qureg: Qureg):
        if qureg.circuit.num_gates > 0:
            raise QuTrunkError("PLUS should be applied at the beginning of circuit.")

        if not isinstance(qureg, Qureg):
            raise TypeError("the operand must be Qureg.")

        self._process_state(qureg)

    def _process_state(self, qureg: Qureg):
        """Process plus state."""
        All(H) * qureg


PLUS = Plus()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM custom gate body."""
from .node import Node


class GateBody(Node):
    """Node for an OPENQASM custom gate body.

    Children is a list of gate operation nodes.
    These are one of barrier, custom_unitary, U, or CX.
    """

    def __init__(self, children):
        """Create the gatebody node."""
        super().__init__("gate_body", children, None)

    def qasm(self):
        """Return the corresponding OPENQASM string."""
        string = ""
        for children in self.children:
            string += "  " + children.qasm() + "\n"
        return string

    def calls(self):
        """Return a list of custom gate names in this gate body."""
        lst = []
        for children in self.children:
            if children.type == "custom_unitary":
                lst.append(children.name)
        return lst
